#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Contains dingo tasks to synchronize the file tree with the ROC database."""

import re

from poppy.core.logger import logger
from poppy.core.db.connector import Connector
from poppy.core.task import Task

from roc.dingo.constants import PIPELINE_DATABASE
from roc.dingo.models.file import FileLog


__all__ = ['UpdateParentsInDb']


class UpdateParentsInDb(Task):
    """
    Update database linking files and their parents, based on the
    error_log field containing "Missing parents"
    """
    plugin_name = 'roc.dingo'
    name = 'update_parents_in_db'

    def add_targets(self):
        pass

    def setup_inputs(self):
        pass

    def remove_parent_from_log(child, parent):
        """
        Take a pair child/parent and remove the child error_log
        part containing the parent message

        :param child: FileLog element for child
        :param parent: FileLog element for parent
        """

        miss_expr = re.compile('Missing parents : (.*)$')

        logger.debug('Child basename : {}'.format(child.basename))
        logger.debug('Child error_log : {}'.format(child.error_log))
        logger.debug('Parent basename : {}'.format(parent.basename))

        # search into the error_log message
        old_missing_parents = miss_expr.search(child.error_log)
        missing_parents = old_missing_parents.group(1).split(', ')

        logger.debug('Child missing parents : {}'.format(missing_parents))

        # remove parent error message
        if parent.basename in missing_parents:
            missing_parents.remove(parent.basename)
        else:
            logger.warning(
                'Parent {} not found in error_log \'{}\'. '
                'This should not happen.'.format(
                    parent.basename,
                    child.error_log
                ))

        # if the parent has .h5 or .cdf as extension
        # try to remove the other extension (both are in the error_log)

        if '.cdf' in parent.basename:
            h5name = parent.basename.replace('.cdf', '.h5')
            if h5name in missing_parents:
                missing_parents.remove(h5name)

        if '.h5' in parent.basename:
            cdfname = parent.basename.replace('.h5', '.cdf')
            if cdfname in missing_parents:
                missing_parents.remove(cdfname)

        # restore the error_log message
        if len(missing_parents) > 0:
            missing_parents = 'Missing parents : ' + \
                ', '.join(missing_parents)
        else:
            missing_parents = ''

        # update error_log in DB
        child.error_log = child.error_log.replace(
            old_missing_parents.group(0),
            missing_parents)

    def run(self):
        logger.debug('UpdateParentsInDb() : run')

        # Get the database connection if needed
        if not hasattr(self, 'session'):
            self.session = Connector.manager[PIPELINE_DATABASE].session

        # Update Database with
        # the pairs (child/parent) which can be linked
        # based on the error_log
        # Returns the pairs inserted

        sql1 = """
            with f1 (id, file, parent) as (select distinct f.id, f.basename ,
                trim(unnest(string_to_array(substring(substring(f.error_log
                    from 'Missing parents : .*$') from 19), ', ')))
                from pipeline.file_log f
                where f.is_removed = False and f.error_log ~ 'Missing parents')
            insert into pipeline.parents_file_log (id_child, id_parent) (
                select f1.id , f2.id
                from f1 inner join pipeline.file_log f2
                on f1.parent = f2.basename
                except (
                    select id_child, id_parent from pipeline.parents_file_log))
            -- on conflict do nothing
            -- to be restored when bdd-lesia will be > 9.4
            returning id_child, id_parent;
            """

        results1 = self.session.execute(sql1)

        cpt = 0
        for row in results1:
            cpt += 1
            file = self.session.query(FileLog).get(row[0])
            parent = self.session.query(FileLog).get(row[1])

            logger.debug(row)
            logger.info('Add parent {:60s} to {:60s} '
                        '[{:05.2f}% complete]'.format(
                            parent.basename,
                            file.basename,
                            100 * cpt / results1.rowcount))

            UpdateParentsInDb.remove_parent_from_log(file, parent)

        # Returns the pairs that could be linked based on the error_log
        # but are already linked in parents_file_log
        # This should not happen, but may be useful after a pipeline error

        sql2 = """
            with f1 (id, file, parent) as (select distinct f.id, f.basename ,
                trim(unnest(string_to_array(substring(substring(f.error_log
                    from 'Missing parents : .*$') from 19), ', ')))
                from pipeline.file_log f
                where f.is_removed = False and f.error_log ~ 'Missing parents')
            select f1.id , f2.id
                from f1 inner join pipeline.file_log f2
                on f1.parent = f2.basename
                intersect (
                    select id_child, id_parent from pipeline.parents_file_log)
            """

        # ensure that preceding instructions have been committed
        self.session.commit()

        results2 = self.session.execute(sql2)

        cpt = 0
        for row in results2:
            cpt += 1
            file = self.session.query(FileLog).get(row[0])
            parent = self.session.query(FileLog).get(row[1])

            logger.debug(row)
            logger.info('Parent {:60s} to {:60s} was already added '
                        '[{:05.2f}% complete]'.format(
                            parent.basename,
                            file.basename,
                            100 * cpt / results2.rowcount))

            UpdateParentsInDb.remove_parent_from_log(file, parent)

        if results1.rowcount + results2.rowcount == 0:
            logger.info('Nothing to be done.')



# -*- coding: utf-8 -*-
"""empty message

Revision ID: roc_dingo_0008_remove_date
Revises: roc_dingo_0007_add_error_log
Create Date: 2021-02-04 16:17:41.195180

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'roc_dingo_0008_remove_date'
down_revision = 'roc_dingo_0007_add_error_log'
branch_labels = None
depends_on = None


def upgrade():
    # Column date is useless.
    # All the information is in start/end time`
    # date column is irrelevant for monthly files.
    op.drop_column('file_log', 'date', schema='pipeline')


def downgrade():
    #
    op.add_column(
        'file_log',
        sa.Column('date', sa.DATE(), nullable=True),
        schema='pipeline')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os

from poppy.core.tools.poppy_argparse import argparse
from poppy.core.command import Command

from roc.dingo.constants import PIPELINE_TABLES, TRYOUTS, \
    TIME_WAIT_SEC, DATA_VERSION, \
    ROC_DATA_ROOT, FILE_LEVEL_LIST, ROC_SPICE_KERNEL_ROOT, SQL_LIMIT, \
    IDB_SOURCE, WORKERS, SOAR_TAP_URL, \
    ROC_DELIVERED_ROOT, ROC_PUBLIC_DATA_ROOT
from roc.dingo.tasks.clear import ClearRocDb
from roc.dingo.tasks.efecs_to_db import EfecsToDb
from roc.dingo.tasks.event_to_db import EventToDb

from roc.dingo.tasks import \
    LogFileToDb, ParseFileTree, CheckFileInDb, UpdateParentsInDb, \
    StoreDelivered, StorePublicFilename, CheckSOARDataAvailability, \
    SearchForUpdate
from roc.dingo.tasks.l0_to_hfrtimelog import L0ToHfrTimeLog
from roc.dingo.tasks.solohk_to_db import SoloHkToDb
from roc.dingo.tasks.db_to_solohk import DbToSoloHk
from roc.dingo.tasks.update_utc_time import UpdateUtcTime
from roc.dingo.tools import valid_data_version, valid_time, valid_date

from roc.dingo.tasks import \
    ExportToJson, L0ToDb, MoveFailedFiles, MoveProcessedFiles

__all__ = []


class DingoCommands(Command):
    """
    Manage the commands relative to the DINGO module.
    """
    __command__ = 'dingo'
    __command_name__ = 'dingo'
    __parent__ = 'master'
    __help__ = """
        Commands relative to the DINGO module, responsible for getting,
        parsing and store de-commuted packets into the ROC database.
    """

    def add_arguments(self, parser):
        """
        Add input arguments common to all the Dingo plugin.

        :param parser: high-level pipeline parser
        :return:
        """

        # If passed as an argument, then generate a temporary file
        # in the output folder
        # To indicate that output file production is in progress
        # lock file is automatically deleted at the end.
        parser.add_argument(
            '--lock-file',
            help='Name of the lock temporary file.',
            default=None,
        )

        # specify the IDB version to use
        parser.add_argument(
            '--idb-version',
            help='IDB version to use.',
            default=[None],
            nargs=1,
        )

        # specify the IDB source to use
        parser.add_argument(
            '--idb-source',
            help='IDB source to use (MIB, SRDB or PALISADE).',
            default=[IDB_SOURCE],
            nargs=1,
        )

        # specify the SOLO SCLK SPICE kernel
        # (required to compute UTC time from OBT time)
        parser.add_argument(
            '--sclk',
            help='Path of SOLO SPICE SCLK kernel file '
                 '(required to compute UTC time from OBT time).',
            default=[None],
            nargs=1,
        )

        # specify the SOLO SPICE LSK kernel
        # (required to compute UTC time from OBT time)
        parser.add_argument(
            '--lsk',
            help='Path of SOLO SPICE LSK kernel '
                 '(required to compute UTC time from OBT time).',
            default=[None],
            nargs=1,
        )

        # Specify the value of the Data_version attribute (and filename)
        parser.add_argument(
            '-v',
            '--data-version',
            nargs=1,
            help='Define the data version value for output files. ',
            type=valid_data_version,
            default=[DATA_VERSION],
        )

        parser.add_argument(
            '-t',
            '--tryouts',
            nargs=1,
            help='Max. number of database query tryouts. ',
            type=int,
            default=[TRYOUTS],
        )

        parser.add_argument(
            '-w',
            '--wait',
            nargs=1,
            help='Time in seconds to wait between two tryouts. ',
            type=int,
            default=[TIME_WAIT_SEC],
        )

        parser.add_argument(
            '-P',
            '--processed-files-dir',
            nargs=1,
            type=str,
            help='Directory where well processed files must be moved'
        )

        parser.add_argument(
            '-F',
            '--failed-files-dir',
            nargs=1,
            type=str,
            help='Directory where failed files must be moved'
        )

        parser.add_argument(
            '--no-move',
            action='store_true',
            default=False,
            help='If passed, skip processed/failed file moving')

        parser.add_argument(
            '-s',
            '--start-time',
            nargs=1, type=str,
            help='Filter database query/insert by start time. ',
            default=[None])

        parser.add_argument(
            '-e',
            '--end-time',
            nargs=1, type=str,
            help='Filter database query/insert by end time. ',
            default=[None])

        parser.add_argument(
            '-l',
            '--limit',
            nargs=1,
            help='Limit of returned rows when querying database. ',
            type=int,
            default=[SQL_LIMIT],
        )

        parser.add_argument(
            '-d',
            '--date-list',
            nargs='+',
            type=valid_date,
            default=[],
            help='List of dates to process (format is YYYYMMDD)')

        parser.add_argument(
            '--to-queue',
            action='store_true',
            default=False,
            help='If passed, add new datasets in the data queue')

        parser.add_argument(
            '--from-queue',
            action='store_true',
            default=False,
            help='If passed, process datasets found in the data queue')

        parser.add_argument(
            '--clean-queue',
            action='store_true',
            default=False,
            help='If passed, clean data queue from already processed datasets')

        parser.add_argument(
            '-W',
            '--workers',
            type=int,
            nargs=1,
            default=[WORKERS],
            help='Number of workers to run in parallel '
                 '(only for multiprocessing tasks)')


class L0ToDbCommand(Command):
    """
    Command to import some packet data from RPW L0 files into the
    ROC database.
    """
    __command__ = 'dingo_l0_to_db'
    __command_name__ = 'l0_to_db'
    __parent__ = 'dingo'
    __parent_arguments__ = ['base']
    __help__ = """
        Command to import some packet data from RPW L0 files
        into the ROC database.
    """

    def add_arguments(self, parser):
        # path to input RPW L0 file(s)
        parser.add_argument(
            '-l0', '--rpw-l0-files',
            help="""
             Input RPW L0 file(s) to import into the ROC database.
             """,
            nargs='+',
            type=str,
            required=True,
        )
        parser.add_argument(
            '-i', '--include',
            nargs='+',
            type=str,
            default=[],
            help='List of packets to include (by palisade_id)',
        )
        parser.add_argument(
            '-x', '--exclude',
            nargs='+',
            type=str,
            default=[],
            help='List of packets to ignore (by palisade_id)',
        )
        parser.add_argument(
            '-T', '--exclude-tm-apid',
            nargs='+',
            type=int,
            default=[],
            help='List of TM packets to exclude (by apid)',
        )
        parser.add_argument(
            '-C', '--exclude-tc-apid',
            nargs='+',
            type=int,
            default=[],
            help='List of TC packets to exclude (by apid)',
        )
        parser.add_argument(
            '--param-only',
            action='store_true',
            default=False,
            help='Insert packet parameters only. '
                 '(i.e., no insertion in tm_log/tc_log tables)'
        )

    def setup_tasks(self, pipeline):
        """
        Run workflow of tasks .
        """

        # Set start task
        start = L0ToDb()
        end = L0ToDb()

        #  Build workflow of tasks
        pipeline | start

        # define the start/end task of the pipeline
        pipeline.start = start
        pipeline.end = end

class L0ToHfrTimeLogCommand(Command):
    """
    Command to import RPW L0 files into the
    pipeline.hfr_time_log table in ROC database.
    """
    __command__ = 'dingo_l0_to_hfrtimelog'
    __command_name__ = 'l0_to_hfrtimelog'
    __parent__ = 'dingo'
    __parent_arguments__ = ['base']
    __help__ = """
         Command to import RPW L0 files into the
         pipeline.hfr_time_log table in ROC database.
    """

    def add_arguments(self, parser):
        # path to input RPW L0 file(s)
        parser.add_argument(
            '-l0', '--rpw-l0-files',
            help="""
             Input RPW L0 file(s) to process.
             """,
            nargs='+',
            type=str,
            required=True,
        )

    def setup_tasks(self, pipeline):
        """
        Run workflow of tasks .
        """

        # Set start task
        start = L0ToHfrTimeLog()
        end = L0ToHfrTimeLog()

        #  Build workflow of tasks
        pipeline | start

        # define the start/end task of the pipeline
        pipeline.start = start
        pipeline.end = end


class EventToDbCommand(Command):
    """
    Command to import some SOLO/RPW-related event data into the ROC database.
    These events are used to fill the QUALITY_BITMASK zVariable in RPW L1 CDF
    """
    __command__ = 'dingo_event_to_db'
    __command_name__ = 'event_to_db'
    __parent__ = 'dingo'
    __parent_arguments__ = ['base']
    __help__ = """
        Command to import some SOLO/RPW-related event data
        into the ROC database.
    """

    def add_arguments(self, parser):
        pass

    def setup_tasks(self, pipeline):
        """
        Run workflow of tasks .
        """

        # Set start task
        start = EventToDb()

        #  Build workflow of tasks
        pipeline | start

        # define the start/end task of the pipeline
        pipeline.start = start


class DbToJsonCommand(Command):
    """
    Command to export ROC pipeline database content in a JSON format file.
    """
    __command__ = 'dingo_db_to_json'
    __command_name__ = 'db_to_json'
    __parent__ = 'dingo'
    __parent_arguments__ = ['base']
    __help__ = """
        Command to export the content of the ROC pipeline database
        in a JSON file.
    """

    def add_arguments(self, parser):

        parser.add_argument(
            '-o', '--output-json',
            help="""
                 Name of the output JSON format file.
                 """,
            type=str,
            default=None,
        )

        parser.add_argument(
            '-t', '--tables',
            help=f"""
                 Name of table(s) to export.
                 Possible values: {','.join(list(PIPELINE_TABLES.keys()))}
                 """,
            nargs='+',
            default=list(PIPELINE_TABLES.keys()),
        )

        parser.add_argument(
            '-t', '--tables',
            help=f"""
                 Name of table(s) to export.
                 Possible values: {','.join(list(PIPELINE_TABLES.keys()))}
                 """,
            nargs='+',
            default=list(PIPELINE_TABLES.keys()),
        )

        parser.add_argument(
            '-C', '--creation-time',
            action='store_true',
            help='Filter time by creation time instead of insertion time '
                 '(only for packet_log and file_log tables'
        )

    def setup_tasks(self, pipeline):
        """
        Export table entries in the ROC pipeline database into JSON file.
        """
        # the task
        start = ExportToJson()

        # build the pipeline workflow for the command
        pipeline | start
        pipeline.start = start


class ClearRocDbCommand(Command):
    """
    Command to clear the content of the ROC pipeline database.
    (Delete pipeline.packet_log and pipeline.file_log tables only)
    """
    __command__ = 'dingo_clear_rocdb'
    __command_name__ = 'clear_rocdb'
    __parent__ = 'dingo'
    __parent_arguments__ = ['base']
    __help__ = """
        Command to clear content of the pipeline
        in the ROC database.
    """

    def add_arguments(self, parser):
        parser.add_argument(
            '-t', '--tables',
            help=f"""
                 Name of table(s) to delete.
                 Possible values: {','.join(list(PIPELINE_TABLES.keys()))}
                 """,
            nargs='+',
            default=list(PIPELINE_TABLES.keys()),
        )

        parser.add_argument(
            '--force',
            help="""
                Force deletion (i.e., no confirmation step)
                """,
            action='store_true',
            default=False,
        )

    def setup_tasks(self, pipeline):
        """
        Clear of table entries in the ROC pipeline database.
        """

        # the task
        clear_rocdb = ClearRocDb()

        # build the pipeline workflow for the command
        pipeline | clear_rocdb
        pipeline.start = clear_rocdb


class LogFileToDbCommand(Command):
    """
    Command to log files from the ROC filetree into the ROC pipeline database
    """
    __command__ = 'dingo_logfile_to_rocdb'
    __command_name__ = 'logfile_to_rocdb'
    __parent__ = 'dingo'
    __parent_arguments__ = ['base']
    __help__ = """
        Command to log files from the ROC filetree
        into the ROC pipeline database.
    """

    def add_arguments(self, parser):

        parser.add_argument(
            '-r', '--root',
            help="""
                 relative path to the ROC filetree root
                 """,
            default=ROC_DATA_ROOT,
            type=valid_data_path_type,
        )

        # for this command, SK is not authorized
        levels = list(FILE_LEVEL_LIST)
        levels.remove('SK')

        parser.add_argument(
            '-l', '--level',
            help=f"""
                 comma-separated list of levels to include in the parsing
                 e.g L1,L2 or HK,L0,BIA
                 Default values: {','.join(list(levels))}
                 """,
            nargs='+',
            choices=list(levels),
            default=list(levels),
        )

        parser.add_argument(
            '-d', '--date',
            help="""
                 date to include in the parsing (yyyy or yyyy/mm or yyyy/mm/dd)
                 e.g 2020 or 2020/07 or 2020/07/02
                 """,
            default=None,
        )

        parser.add_argument(
            '-f', '--force',
            help="""
                 Force the update of the database
                 """,
            action='store_true',
            default=False,
        )

        parser.add_argument(
            '-nf', '--no-former-versions',
            help="""
                 Exclude files from the directories "former_versions"
                 """,
            action='store_true',
            default=False,
        )

    def setup_tasks(self, pipeline):
        """
        Synchronise the ROC file tree content into the ROC database
        """

        # the tasks
        parse_file_tree = ParseFileTree()
        log_file_to_db = LogFileToDb()
        check_file_in_db = CheckFileInDb()

        # build the pipeline workflow for the command
        pipeline | parse_file_tree | check_file_in_db | log_file_to_db

        # define the start points of the pipeline
        pipeline.start = parse_file_tree
        pipeline.end = log_file_to_db


class UpdateLogFileCommand(Command):
    """
    Search for files to be updated in file file_log table
    Their column to_update is set to True
    Update these files (set to_update=False)
    """
    __command__ = 'dingo_update_filelog'
    __command_name__ = 'update_filelog'
    __parent__ = 'dingo'
    __parent_arguments__ = ['base']
    __help__ = """
        Search for files to be updated in file file_log table
        and update them
    """

    def add_arguments(self, parser):

        parser.add_argument(
            '-r', '--root',
            help="""
                 relative path to the ROC filetree root
                 """,
            default=ROC_DATA_ROOT,
            type=valid_data_path_type,
        )

    def setup_tasks(self, pipeline):
        """
        Update the requested files (to_update = True in db)
        """

        # the tasks
        search_for_update = SearchForUpdate()
        log_file_to_db = LogFileToDb()

        # build the pipeline workflow for the command
        pipeline | search_for_update | log_file_to_db

        # define the start points of the pipeline
        pipeline.start = search_for_update
        pipeline.end = log_file_to_db


class LogSpiceKernelsToDbCommand(Command):
    """
    Command to log files from the ROC filetree into the ROC pipeline database
    """
    __command__ = 'dingo_logsk_to_rocdb'
    __command_name__ = 'logsk_to_rocdb'
    __parent__ = 'dingo'
    __parent_arguments__ = ['base']
    __help__ = """
        Command to log spice kernel files from the ROC filetree
        into the ROC pipeline database.
    """

    def add_arguments(self, parser):

        parser.add_argument(
            '-r', '--root',
            help="""
                 relative path to the ROC filetree root
                 """,
            type=valid_data_path_type,
            default=ROC_SPICE_KERNEL_ROOT
        )

        parser.add_argument(
            '-l', '--level',
            help="""
                 level to include in the parsing
                 Only one value possible : SK
                 """,
            nargs='+',
            choices=['SK'],
            default=['SK'],
        )

        parser.add_argument(
            '-f', '--force',
            help="""
                 Force the update of the database
                 """,
            action='store_true',
            default=False,
        )

        parser.add_argument(
            '-nf', '--no-former-versions',
            help="""
                 Exclude files from the directories "former_versions"
                 """,
            action='store_true',
            default=False,
        )

    def setup_tasks(self, pipeline):
        """
        Synchronise the ROC file tree content into the ROC database
        """

        # the tasks
        parse_file_tree = ParseFileTree()
        log_file_to_db = LogFileToDb()
        check_file_in_db = CheckFileInDb()

        # build the pipeline workflow for the command
        pipeline | parse_file_tree | check_file_in_db | log_file_to_db

        # define the start points of the pipeline
        pipeline.start = parse_file_tree
        pipeline.end = log_file_to_db


class UpdateMissingParentsDbCommand(Command):
    """
    Command to log files from the ROC filetree into the ROC pipeline database
    """
    __command__ = 'dingo_update_missing_parents_in_db'
    __command_name__ = 'update_missing_parents_in_db'
    __parent__ = 'dingo'
    __parent_arguments__ = ['base']
    __help__ = """
        Command to update files into the ROC pipeline database.
        Only handles files with the error_log Missing Parents.
    """

    def add_arguments(self, parser):
        pass

    def setup_tasks(self, pipeline):
        """
        Synchronise the ROC file tree content into the ROC database
        """

        # the tasks
        update_parents_in_db = UpdateParentsInDb()

        # build the pipeline workflow for the command
        pipeline | update_parents_in_db

        # define the start points of the pipeline
        pipeline.start = update_parents_in_db
        pipeline.end = update_parents_in_db


class StoreDeliveredCommand(Command):
    """
    Command to store into file_log files that have been delivered by GFTS
    """
    __command__ = 'dingo_store_delivered'
    __command_name__ = 'store_delivered'
    __parent__ = 'dingo'
    __parent_arguments__ = ['base']
    __help__ = """
        Command to store into file_log files that have been delivered by GFTS.
    """

    def add_arguments(self, parser):

        parser.add_argument(
            '--delivered',
            help="""
                 relative path to the ROC delivered directory
                 """,
            type=valid_data_path_type,
            default=ROC_DELIVERED_ROOT
        )

        parser.add_argument(
            '--public',
            help="""
                 relative path to the ROC delivered directory
                 """,
            type=valid_data_path_type,
            default=ROC_PUBLIC_DATA_ROOT
        )

        parser.add_argument(
            '--remove_files',
            help="""
                 If set shadow delivered files will be removed
                 """,
            action='store_true',
            default=False,
        )

    def setup_tasks(self, pipeline):
        """
        Synchronise the ROC file tree content into the ROC database
        """

        # the tasks
        store_public = StorePublicFilename()
        store_delivered = StoreDelivered()

        # build the pipeline workflow for the command
        pipeline | store_public | store_delivered

        # define the start points of the pipeline
        pipeline.start = store_public
        pipeline.end = store_delivered


class CheckSOARDataAvailabilityCommand(Command):
    """
    Query and report RPW data availability in Solar Orbiter Archive (SOAR)
    at ESAC.
    https://gitlab.obspm.fr/ROC/Pipelines/RODP/-/blob/develop/scripts/report_rpw_soar_data_availability.py  # noqa: E501
    """
    __command__ = 'dingo_check_soar'
    __command_name__ = 'check_soar'
    __parent__ = 'dingo'
    __parent_arguments__ = ['base']
    __help__ = """
        Check if delivered files are available in SOAR.
    """

    def add_arguments(self, parser):
        parser.add_argument(
            '--url',
            type=str,
            default=SOAR_TAP_URL,
            help='URL of the SOAR TAP service. '
            f'{SOAR_TAP_URL}'
        )

    def setup_tasks(self, pipeline):
        """
        Check if delivered files are available in SOAR
        """

        # the tasks
        check_soar = CheckSOARDataAvailability()

        # build the pipeline workflow for the command
        pipeline | check_soar

        # define the start points of the pipeline
        pipeline.start = check_soar
        pipeline.end = check_soar


def valid_data_path_type(arg):
    """ Type function for argparse - an accessible path not empty """

    # check is accessible
    ret = os.access(arg, os.R_OK)
    if not ret:
        raise argparse.ArgumentTypeError(
            'Argument must be a valid and readable data path')

    # check if not empty
    listdir = os.listdir(arg)
    if len(listdir) == 0:
        raise argparse.ArgumentTypeError(
            'Argument data path must contain data. Directory is empty.')

    return arg


class SoloHkToDbCommand(Command):
    """
    Command to import Solo HK EDDS data into the ROC database.
    """
    __command__ = 'dingo_solohk_to_db'
    __command_name__ = 'solohk_to_db'
    __parent__ = 'dingo'
    __parent_arguments__ = ['base']
    __help__ = """
        Command to import Solo HK EDDS data into the ROC database.
    """

    def add_arguments(self, parser):
        # path to input RPW L0 file(s)
        parser.add_argument(
            '-f', '--solo-hk-files',
            help="""
             Input SOLO HK EDDS XML file(s) to import into the ROC database.
             """,
            nargs='+',
            type=str,
            required=True,
        )

    def setup_tasks(self, pipeline):
        """
        Insert SOLO HK EDDS Data into the ROC database
        """

        # Task
        start_task = SoloHkToDb()
        end_task = MoveFailedFiles()

        # build the pipeline workflow for the command
        pipeline | start_task | MoveProcessedFiles() | end_task

        # define the start points of the pipeline
        pipeline.start = start_task
        pipeline.end = end_task


class DbToSoloHkCmd(Command):
    """
    Command to export Solo HK EDDS data saved in the ROC database
    into a output XML file.
    Format of the output XML file is close to the Parameter EDDS response XML.
    """
    __command__ = 'dingo_db_to_solohk'
    __command_name__ = 'db_to_solohk'
    __parent__ = 'dingo'
    __parent_arguments__ = ['base']
    __help__ = """
        Command to export Solo HK EDDS data saved in the ROC database
        into a output XML file.
    """

    def add_arguments(self, parser):
        pass

    def setup_tasks(self, pipeline):
        """
        Export SOLO HK EDDS Data into an output XML file
        """

        # Task
        start_task = DbToSoloHk()
        end_task = MoveFailedFiles()

        # build the pipeline workflow for the command
        pipeline | start_task | end_task

        # define the start points of the pipeline
        pipeline.start = start_task
        pipeline.end = end_task


class EfecsToDbCommand(Command):
    """
    Command to import EFECS data into the ROC database.
    """
    __command__ = 'dingo_efecs_to_db'
    __command_name__ = 'efecs_to_db'
    __parent__ = 'dingo'
    __parent_arguments__ = ['base']
    __help__ = """
        Command to import Solo EFECS data into the ROC database.
    """

    def add_arguments(self, parser):
        # path to input RPW L0 file(s)
        parser.add_argument(
            '-f', '--efecs-files',
            help="""
             Input SOLO EFECS XML file(s) to import into the ROC database.
             """,
            nargs='+',
            type=str,
            required=True,
        )

    def setup_tasks(self, pipeline):
        """
        Insert SOLO HK EFECS Data into the ROC database
        """

        # Task
        start_task = EfecsToDb()

        # build the pipeline workflow for the command
        pipeline | start_task

        # define the start points of the pipeline
        pipeline.start = start_task


class UpdateUtcTimeCommand(Command):
    """
    Command to update utc_time values in the ROC database.
    """
    __command__ = 'dingo_update_utc_time'
    __command_name__ = 'update_utc_time'
    __parent__ = 'dingo'
    __parent_arguments__ = ['base']
    __help__ = """
        Command to update utc_time values in the ROC database.
    """

    def add_arguments(self, parser):
        pass

    def setup_tasks(self, pipeline):
        """
        Update utc_time values in the ROC database
        """

        # Task
        start_task = UpdateUtcTime()

        # build the pipeline workflow for the command
        pipeline | start_task

        # define the start points of the pipeline
        pipeline.start = start_task



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
from datetime import datetime, timedelta
import json
import os.path as osp

from jinja2 import Environment, FileSystemLoader

from poppy.core.tools.exceptions import DescriptorLoadError
from poppy.core.logger import logger
from poppy.core.conf import settings

__all__ = ['PLUGIN',
           'DESCRIPTOR',
           'NAIF_SOLO_ID',
           'TM_PACKET_TYPE',
           'TC_PACKET_TYPE',
           'PACKET_TYPE',
           'PACKET_DATA_GROUP',
           'PACKET_INVALID',
           'PACKET_VALID',
           'IDB_SOURCE',
           'TEST_DATABASE',
           'PIPELINE_DATABASE',
           'PIPELINE_TABLES',
           'SQL_LIMIT',
           'TRYOUTS',
           'TIME_SQL_STRFORMAT',
           'TIME_INPUT_STRFORMAT',
           'TIME_OUTPUT_STRFORMAT',
           'TIME_DAILY_STRFORMAT',
           'TIME_ISO_STRFORMAT',
           'TIME_EFECS_STRFORMAT',
           'TIME_WAIT_SEC',
           'DATA_VERSION',
           'START_TIME',
           'END_TIME',
           'PACKET_CHUNK',
           'DELTA_HOURS',
           'TC_ACK_ALLOWED_STATUS',
           'ROC_DATA_ROOT',
           'ROC_PUBLIC_DATA_ROOT',
           'FILE_STATE_LIST',
           'FILE_STATUS_LIST',
           'FILE_LEVEL_LIST',
           'ROC_SPICE_KERNEL_ROOT',
           'DATA_ALLOWED_EXTENSIONS',
           'SPICE_KERNEL_ALLOWED_EXTENSIONS',
           'LEVEL_2_INST',
           'LEVEL_3_INST',
           'LEVEL_BY_MISSION',
           'LEVEL_BY_MONTH',
           'LEVEL_BY_DAY',
           'SPICE_KERNEL_TYPES',
           'ROC_DATA_HTTPS_ROOT',
           'ROC_DATA_HTTPS_PUBLIC_ROOT',
           'L0_DIRPATH_PATTERN',
           'ROC_SPICE_KERNEL_HTTPS_ROOT',
           'BIA_SWEEP_TABLE_PACKETS',
           'HFR_FREQ_LIST_PACKETS',
           'SBM_LOG_PACKETS',
           'HFR_SCIENCE_PACKETS',
           'BIA_SWEEP_LOG_PACKETS',
           'SBM_ALGO_PARAM_LIST',
           'BIA_SWEEP_TABLE_NR',
           'LFR_KCOEFF_PARAM_NR',
           'JENV',
           'SOLO_HK_TEMPLATE',
           'BIA_CURRENT_LOG_PACKETS',
           'LFR_KCOEFF_DUMP_PACKETS',
           'TF_CP_BIA_P011_SRDB_ID',
           'WORKERS',
           'CIWT0130TM',
           'TF_PA_DPU_0038',
           'TF_PA_DPU_0039',
           'SOAR_TAP_URL',
           'EFECS_EVENT_LOG',
           'EVENT_TM_APID',
           'TC_ACK_APID',
           'SCIENCE_TM_APID',
           'LL_TM_APID',
           'SBM_TM_APID',
           'HK_TM_APID',
           'TM_APID_EVENT_LOG',
           'TC_PACKET_EVENT_LOG',
           'LFR_CALIB_EVENT_LOG',
           ]

# Name of the plugin
from roc.dingo.models.file import FileLog
from roc.dingo.models.file import \
    FILE_STATE_LIST, FILE_STATUS_LIST, FILE_LEVEL_LIST
from roc.dingo.models.packet import TmLog, TcLog

PLUGIN = 'roc.dingo'

# root directory of the module
_ROOT_DIRECTORY = osp.abspath(
    osp.join(
        osp.dirname(__file__),
    )
)


# NAIF SPICE ID for SOLO
NAIF_SOLO_ID = -144

# Default data version for product file
DATA_VERSION = '01'

# Number of database connexion tryouts
TRYOUTS = 3

# Time to wait in seconds between two database connection tryouts
TIME_WAIT_SEC = 3

# Default IDB source
IDB_SOURCE = 'MIB'

# TM/TC packet_type CCSDS ID
TM_PACKET_TYPE = 0
TC_PACKET_TYPE = 1
PACKET_VALID = 'VALID'
PACKET_INVALID = 'INVALID'

# Packet type list
PACKET_TYPE = ['TM', 'TC']

# Group for TM and TC
PACKET_DATA_GROUP = {'TM': 'source_data',
                     'TC': 'application_data'}

# Time string format
TIME_SQL_STRFORMAT = '%Y-%m-%dT%H:%M:%S.%f'
TIME_ISO_STRFORMAT = '%Y-%m-%dT%H:%M:%S.%fZ'
TIME_INPUT_STRFORMAT = '%Y-%m-%dT%H:%M:%S'
TIME_EFECS_STRFORMAT = '%Y-%jT%H:%M:%SZ'
TIME_OUTPUT_STRFORMAT = '%Y%m%dT%H%M%S'
TIME_DAILY_STRFORMAT = '%Y%m%d'


# Limit of rows to be returned by the database
SQL_LIMIT = 1000000000

# Number of workers to run in parallel
WORKERS = 1

# Pipeline tables
PIPELINE_TABLES = {'tm_log': TmLog,
                   'file_log': FileLog,
                   'tc_log': TcLog,
                   }

# Default end time of the packets time range to query
# Default is beginning of the current day
END_TIME = datetime.now().date()

# Default start time of the packets time range to query
# Default is end_time - 180 days
START_TIME = END_TIME - timedelta(days=180)

# Load pipeline database identifier
try:
    PIPELINE_DATABASE = settings.PIPELINE_DATABASE
except Exception:
    PIPELINE_DATABASE = 'PIPELINE_DATABASE'
    logger.warning(f'settings.PIPELINE_DATABASE not defined for {__file__}, \
                     use "{PIPELINE_DATABASE}" by default!')

try:
    TEST_DATABASE = settings.TEST_DATABASE
except Exception:
    TEST_DATABASE = 'TEST_DATABASE'
    logger.warning(f'settings.TEST_DATABASE not defined for {__file__}, \
                     use "{TEST_DATABASE}" by default!')

# Load descriptor file
descriptor_path = osp.join(_ROOT_DIRECTORY, 'descriptor.json')
try:
    with open(descriptor_path, 'r') as file_buffer:
        DESCRIPTOR = json.load(file_buffer)
except Exception:
    raise DescriptorLoadError(f'Loading {descriptor_path} has failed!')

# Number of packets to process/insert in one chunk
PACKET_CHUNK = 100

# Time interval in hours to query
DELTA_HOURS = 1

# TC Packet ack possible status
TC_ACK_ALLOWED_STATUS = ['PASSED', 'FAILED']

# ROC data file tree rootname
ROC_DATA_ROOT = \
    '/volumes/plasma/rpw/roc/data/https/private/solo/rpw/data/'

# ROC publica data file tree rootname
ROC_PUBLIC_DATA_ROOT = \
    '/volumes/plasma/rpw/roc/data/https/pub/solo/rpw/data/'

# L0 FILE directory PATH PATTERN (default path in pipeline Docker container)
L0_DIRPATH_PATTERN = '/pipeline/data/private/solo/rpw/data/L0/????/??/??'

# ROC data http access URL
ROC_DATA_HTTPS_ROOT = \
    'https://rpw.lesia.obspm.fr/roc/data/private/solo/rpw/data/'

# ROC spice kernel file tree rootname
ROC_SPICE_KERNEL_ROOT = \
    '/volumes/plasma/rpw/roc/data/https/private/solo/soc/spice/kernels/'

# ROC data http access URL
ROC_SPICE_KERNEL_HTTPS_ROOT = \
    'https://rpw.lesia.obspm.fr/roc/data/private/solo/soc/spice/kernels/'

# ROC public data http access URL
ROC_DATA_HTTPS_PUBLIC_ROOT = \
    'https://rpw.lesia.obspm.fr/roc/data/pub/solo/rpw/data/'

# ROC Delivered directory
ROC_DELIVERED_ROOT = '/volumes/plasma/rpw/roc/data/sftp/soc/gfts/delivered'

# DATA ALLOWED EXTENSIONS
DATA_ALLOWED_EXTENSIONS = \
    ['xml', 'json', 'cdf', 'h5']

# SPICE KERNEL ALLOWED EXTENSIONS
SPICE_KERNEL_ALLOWED_EXTENSIONS = \
    ['bc', 'bpc', 'bsp', 'tf', 'ti', 'tls', 'tm', 'tpc', 'tsc']

# Defined in models/file.py
# FILE_LEVEL_LIST = [
#     'CAL', 'TC', 'TM', 'SK', 'L0', 'HK', 'L1',
#     'BIA', 'L1R', 'L2', 'L3', 'LL01']

# LEVEL 2 DATA INSTRUMENTS
LEVEL_2_INST = \
    ['lfr_asm', 'lfr_bp', 'lfr_wf_b', 'lfr_wf_e', 'tds_hist1d', 'tds_hist2d',
     'tds_lfm', 'tds_mamp', 'tds_stat', 'tds_wf_b', 'tds_wf_e', 'thr']

# LEVEL 3 DATA INSTRUMENTS
LEVEL_3_INST = ['lfr_density', 'lfr_efield', 'lfr_scpot', 'tnr_fp']

# SPICE KERNEL TYPES
# Skipping mk/ because they are only meta-kernels
SPICE_KERNEL_TYPES = ['ck', 'fk', 'ik', 'lsk', 'pck', 'sclk', 'spk']

# GROUPING PERIOD PER LEVEL
LEVEL_BY_MISSION = ['CAL', 'SK']
LEVEL_BY_MONTH = ['L2', 'L3', 'BIA']
LEVEL_BY_DAY = ['TC', 'TM', 'L0', 'HK', 'L1', 'L1R', 'LL01']


# TM packet APID
TC_ACK_APID = [1201, 1207, 1233, 1249]  # DPU, LFR, TDS, THR
EVENT_TM_APID = [1207]
LL_TM_APID = [1212]
SCIENCE_TM_APID = [1228, 1244, 1260]
SBM_TM_APID = [1276, 1292]
HK_TM_APID = [1204, 1220, 1236, 1252, 1269, 1300]

# List of TM/TC packets to insert in specific tables

SBM_LOG_PACKETS = ['TM_DPU_EVENT_PR_DPU_SBM1',
                   'TM_DPU_EVENT_PR_DPU_SBM2']
BIA_SWEEP_TABLE_PACKETS = ['TC_DPU_LOAD_BIAS_SWEEP',
                           'TC_DPU_CLEAR_BIAS_SWEEP']
BIA_CURRENT_LOG_PACKETS = ['TC_DPU_SET_BIAS1',
                           'TC_DPU_SET_BIAS2',
                           'TC_DPU_SET_BIAS3']
BIA_SWEEP_LOG_PACKETS = ['TM_DPU_EVENT_PR_BIA_SWEEP']
HFR_FREQ_LIST_PACKETS = ['TC_THR_LOAD_NORMAL_PAR_2',
                         'TC_THR_LOAD_NORMAL_PAR_3',
                         'TC_THR_LOAD_BURST_PAR_2',
                         'TC_THR_LOAD_BURST_PAR_3']
LFR_KCOEFF_DUMP_PACKETS = ['TM_LFR_KCOEFFICIENTS_DUMP']
HFR_SCIENCE_PACKETS = ['TM_THR_SCIENCE_NORMAL_HFR',
                       'TM_THR_SCIENCE_BURST_HFR',
                       'TM_THR_SCIENCE_CALIBRATION_HFR']

# SBM1/SBM2 TM algo parameters
SBM_ALGO_PARAM_LIST = {
    1: ['SY_DPU_SBM1_DT1_SBM1_D',
        'SY_DPU_SBM1_DT2_SBM1_D',
        'SY_DPU_SBM1_DT3_SBM1_D'],
    2: ['HK_DPU_SBM2_DT_SBM2',
        'HK_DPU_SBM2_DT_LW',
        'HK_DPU_EPD_S20_ESW_FLAG_D',
        'HK_DPU_EPD_S20_EASW_FLAG_D',
        'HK_DPU_EPD_S20_EN_FLAG_D',
        'HK_DPU_EPD_S20_ES_FLAG_D',
        'HK_DPU_EPD_S20_PSW_FLAG_D',
        'HK_DPU_EPD_S20_PASW_FLAG_D',
        'HK_DPU_EPD_S20_PN_FLAG_D',
        'HK_DPU_EPD_S20_PS_FLAG_D']
}

# List of EFECS events to copy in event_log table
# See EFECS_ICD_SOL-SGS-ICD-0006 for list of EFECS
EFECS_EVENT_LOG = ['WOL', 'TCM', 'PASS', 'EMC_MAND_QUIET',
                   'EMC_PREF_NOISY', 'MAINT', 'NAV']

# List of TM APID to copy in event_log table
TM_APID_EVENT_LOG = [1207]

# List of TC to copy in event_log table
TC_PACKET_EVENT_LOG = ['TC_DPU_SWITCH_ON_EQUIPMENT',
                       'TC_DPU_SWITCH_OFF_EQUIPMENT',
                       'TC_DPU_BOOT_DAS',
                       'TC_DPU_ENTER_STANDBY',
                       'TC_DPU_ENTER_SERVICE',
                       'TC_DPU_ENTER_SURVEY_NORMAL',
                       'TC_DPU_ENTER_SURVEY_BURST',
                       'TC_DPU_ENTER_SURVEY_BACKUP',
                       'TC_DPU_ENTER_SBM_DETECTION',
                       'TC_DPU_ENTER_SBM1_DUMP',
                       'TC_DPU_ENTER_SBM2_ACQ',
                       'TC_TDS_DUMP_SBM2_TSWF',
                       'TC_TDS_DUMP_NORMAL_TSWF',
                       'TC_THR_ENABLE_CALIBRATION',
                       ]
TC_PACKET_EVENT_LOG.extend(HFR_FREQ_LIST_PACKETS)
TC_PACKET_EVENT_LOG.extend(BIA_CURRENT_LOG_PACKETS)

# LFR Calibration event packets
LFR_CALIB_EVENT_LOG = ['TC_LFR_ENABLE_CALIBRATION',
                       'TC_LFR_DISABLE_CALIBRATION']

# Number of max. values that can be stored in the bias sweep table
BIA_SWEEP_TABLE_NR = 256

# Number of kCoeff parameters by TM
LFR_KCOEFF_PARAM_NR = 32

# Setup jinja2 environment
JINJA_TEMPLATE_DIR = os.path.join(_ROOT_DIRECTORY, 'templates')
JENV = Environment(loader=FileSystemLoader(JINJA_TEMPLATE_DIR))
# Jinja2 template for SOLO HK XML files
SOLO_HK_TEMPLATE = 'solo_hk_xml.tpl'

# SRDB ID of the transfer function TF_CP_BIA_P011 (or TF_CP_BIA_0011 in ICD)
# NOTES: SRDB_ID is CIWP0075TC and not CIWP0040TC as indicated in the ICD)
TF_CP_BIA_P011_SRDB_ID = 'CIWP0075TC'

# SBM1 QF TF ID
TF_PA_DPU_0038 = 'CIWP0028TM'
# SBM2 QF TF ID
TF_PA_DPU_0039 = 'CIWP0029TM'

# Enumeration for PA_DPU_BIA_SWEEP_PR_CODE {NIW01757}
CIWT0130TM = {
    1: 'START_ANT1',
    2: 'END_ANT1',
    3: 'START_ANT2',
    4: 'END_ANT2',
    5: 'START_ANT3',
    6: 'END_ANT3',
    7: 'STEP_ANT1',
    8: 'STEP_ANT2',
    9: 'STEP_ANT3',
}

# URL of the SOAR TAP service
SOAR_TAP_URL = 'http://soar.esac.esa.int/soar-sl-tap/tap'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Contains dingo tasks to synchronize the file tree with the ROC database."""

import os

from sqlalchemy import and_
from sqlalchemy.dialects import postgresql
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.orm import load_only

from poppy.core.logger import logger
from poppy.core.db.connector import Connector
from poppy.core.task import Task
from poppy.core.target import BaseTarget

from roc.dingo.constants import PIPELINE_DATABASE, ROC_DELIVERED_ROOT
from roc.dingo.models.file import FileLog


__all__ = ['StoreDelivered']


class StoreDelivered(Task):
    """
    Parse ROC `delivered` folder, and store into database that all these
    files have been successfully delivered to ESA (column is_delivered)
    """
    plugin_name = 'roc.dingo'
    name = 'store_delivered'

    def add_targets(self):
        logger.debug('StoreDelivered() : add_targets')

        self.add_output(target_class=BaseTarget,
                        many=True,
                        identifier='store_delivered_results')

    def setup_inputs(self):
        """
        Setup task inputs.

        :param task:
        :return:
        """

        logger.debug('StoreDelivered() : setup_inputs')

        # get the root file tree
        self.root = self.pipeline.get(
            'delivered', default=ROC_DELIVERED_ROOT, args=True)

        # ensure that there is / at the end
        self.root = os.path.join(self.root, '')

        # get the remove_files flag
        self.remove_files = self.pipeline.get(
            'remove_files', default=False, args=True)

    def run(self):

        logger.debug('StoreDelivered() : run')

        # Initialize inputs
        self.setup_inputs()

        # Initialize counters
        cpt_remove = 0
        cpt_to_store = 0
        nb_files_stored = 0

        # Get the database connection if needed
        if not hasattr(self, 'session'):
            self.session = Connector.manager[PIPELINE_DATABASE].session

        # Get the file list
        to_store = []
        files = os.listdir(self.root)
        for file in files:
            if os.path.isfile(os.path.join(self.root, file)):
                to_store.append(file)

        logger.info(f'{len(to_store)} files in DELIVERED root')

        # Find corresponding entries in file_log

        name_filter = FileLog.public_filename.in_(to_store)
        is_not_delivered_filter = FileLog.is_delivered == False   # noqa: E712
        fields = ['id', 'basename', 'public_filename']

        try:
            query = self.session.query(FileLog).\
                options(load_only(*fields)).\
                filter(and_(name_filter, is_not_delivered_filter))
            results = query.all()
        except NoResultFound:
            logger.info('Nothing to be done')
        except Exception:
            logger.error('Query failed :')
            logger.error(
                str(query.statement.compile(
                    dialect=postgresql.dialect(),
                    compile_kwargs={'literal_binds': True})))
        else:
            logger.info(f'{len(results)} files found in DB')
            cpt_to_store = len(results)

            item_dict = {}
            item_dict['is_delivered'] = True
            try:
                nb_files_stored = self.session.query(FileLog).\
                    filter(and_(name_filter, is_not_delivered_filter)).\
                    update(item_dict, synchronize_session=False)

            except Exception as e:
                logger.error(f'is_delivered update has failed: \n {e}')
                logger.error(
                    str(query.statement.compile(
                        dialect=postgresql.dialect(),
                        compile_kwargs={'literal_binds': True})))
            else:
                if nb_files_stored == cpt_to_store:
                    logger.info(
                        f'{nb_files_stored} files have been updated with '
                        'is_delievered = True')
                else:
                    logger.warning(
                        f'{nb_files_stored}/{cpt_to_store} files have been '
                        'updated with is_delievered = True')

            if self.remove_files:
                # Now, delete file in DELIVERED folder
                for r in results:
                    file_to_delete = os.path.join(self.root, r.public_filename)
                    if os.path.isfile(file_to_delete):
                        try:
                            os.remove(file_to_delete)
                        except FileNotFoundError as e:
                            logger.error(
                                'Unable to remove '
                                f'{r.public_filename}: \n {e}')
                        else:
                            cpt_remove += 1
                            logger.debug(f'Remove {r.public_filename}')

                if cpt_remove != cpt_to_store:
                    logger.warning(
                        f'{cpt_remove}/{cpt_to_store} shadow '
                        'delivered files have been removed')
                else:
                    logger.info(
                        f'{cpt_remove} shadow delivered files '
                        'have been removed')

        # number of stored files
        self.outputs['store_delivered_results'].data = {
            'files_to_be_stored': cpt_to_store,
            'files_stored': nb_files_stored,
            'removed_files': cpt_remove
        }



# -*- coding: utf-8 -*-
"""empty message

Revision ID: roc_dingo_0025_add_table_events
Revises: roc_dingo_0025_add_views_for_anomalies
Create Date: 2022-11-14 17:02:37.168657

"""
from alembic import op
import sqlalchemy as sa
from pathlib import Path
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = 'roc_dingo_0026_add_table_events'
down_revision = 'roc_dingo_0025_add_views_for_anomalies'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'events',
        sa.Column('id', sa.BIGINT(), nullable=False),
        sa.Column('label', sa.String(), nullable=False),
        sa.Column('is_tracked', sa.BOOLEAN(), nullable=False),
        sa.Column('is_anomaly', sa.BOOLEAN(), nullable=False),
        sa.Column('origin', postgresql.ENUM(
            'SOLO', 'RPW', name='event_origin_enum'), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('label'),
        schema='pipeline'
    )

    # set default values
    op.alter_column('events', 'is_tracked',
                    server_default=sa.schema.DefaultClause('1'),  # true
                    schema='pipeline')
    op.alter_column('events', 'is_anomaly',
                    server_default=sa.schema.DefaultClause('0'),  # false
                    schema='pipeline')

    op.execute('DROP VIEW IF EXISTS pipeline.anomalies_per_day')
    op.drop_table('anomalies', schema='pipeline')

    # insert data into events
    filename = 'SQL/roc_dingo_0026_insert_into_events.sql'
    sql_insert_file = Path(__file__).parent / filename
    with open(sql_insert_file, 'r') as sql_file:
        sql_cmd = sql_file.read()
        op.execute(sql_cmd)

    # create views
    filename = 'SQL/roc_dingo_0026_create_view_anomalies_per_day.sql'
    sql_file = Path(__file__).parent / filename
    with open(sql_file, 'r') as sql_content:
        sql_cmd = sql_content.read()
        op.execute(sql_cmd)

    filename = 'SQL/roc_dingo_0026_create_view_rpw_events_per_day.sql'
    sql_file = Path(__file__).parent / filename
    with open(sql_file, 'r') as sql_content:
        sql_cmd = sql_content.read()
        op.execute(sql_cmd)


def downgrade():

    # anomalies_per_day depends on events, delete it first
    op.execute('DROP VIEW IF EXISTS pipeline.anomalies_per_day')
    op.execute('DROP VIEW IF EXISTS pipeline.rpw_events_per_day')
    op.drop_table('events', schema='pipeline')

    # drop enumeration
    op.execute('DROP TYPE IF EXISTS event_origin_enum;')

    # Re create anomalies
    op.create_table('anomalies',
                    sa.Column('id', sa.BIGINT(), nullable=False),
                    sa.Column('label', sa.String(), nullable=False),
                    sa.Column('tracked', sa.BOOLEAN(), nullable=False),
                    sa.PrimaryKeyConstraint('id'),
                    sa.UniqueConstraint('label'),
                    schema='pipeline')

    # set default value to TRUE
    op.alter_column('anomalies', 'tracked',
                    server_default=sa.schema.DefaultClause('1'),
                    schema='pipeline')

    # insert data into anomalies
    filename = 'SQL/roc_dingo_0025_insert_into_anomalies.sql'
    sql_insert_file = Path(__file__).parent / filename
    with open(sql_insert_file, 'r') as sql_file:
        sql_cmd = sql_file.read()
        op.execute(sql_cmd)

    # recreate view anomalies_per_day
    filename = 'SQL/roc_dingo_0025_create_view_anomalies_per_day.sql'
    sql_file = Path(__file__).parent / filename
    with open(sql_file, 'r') as sql_content:
        sql_cmd = sql_content.read()
        op.execute(sql_cmd)



"""empty message

Revision ID: roc_dingo_0012_add_data
Revises: roc_dingo_0011_update_packet
Create Date: 2021-07-15 14:33:41.549410

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

from poppy.pop.alembic.helpers import \
    create_table, create_schema, execute, user, drop_table

# revision identifiers, used by Alembic.
revision = 'roc_dingo_0012_add_data'
down_revision = 'roc_dingo_0011_update_packet'
branch_labels = None
depends_on = None


def upgrade():

    # delete the types if they exist
    execute('DROP TYPE IF EXISTS sbm_status_type')

    # ### commands auto generated by Alembic - please adjust! ###
    create_table('bia_sweep_table',
                 sa.Column('id', sa.BIGINT(), nullable=False),
                 sa.Column('tc_name', sa.String(length=256), nullable=False),
                 sa.Column('utc_time', postgresql.TIMESTAMP(), nullable=False),
                 sa.Column('sweep_eeprom', sa.BOOLEAN(), nullable=True),
                 sa.Column('sweep_step_cur', sa.String(), nullable=True),
                 sa.Column('insert_time', postgresql.TIMESTAMP(),
                           nullable=True),
                 sa.PrimaryKeyConstraint('id'),
                 sa.UniqueConstraint('utc_time', 'tc_name'),
                 schema='pipeline'
                 )
    create_table('bia_sweep_log',
                 sa.Column('id', sa.BIGINT(), nullable=False),
                 sa.Column('sweep_step', sa.String(length=16), nullable=False),
                 sa.Column('utc_time', postgresql.TIMESTAMP(), nullable=False),
                 sa.Column('cuc_time', sa.String(length=1024), nullable=False),
                 sa.Column('utc_time_is_predictive',
                           sa.BOOLEAN(), nullable=True),
                 sa.Column('insert_time', postgresql.TIMESTAMP(),
                           nullable=True),
                 sa.PrimaryKeyConstraint('id'),
                 sa.UniqueConstraint('sweep_step', 'utc_time'),
                 schema='pipeline'
                 )
    create_table('hfr_freq_list',
                 sa.Column('id', sa.BIGINT(), nullable=False),
                 sa.Column('utc_time', postgresql.TIMESTAMP(), nullable=False),
                 sa.Column('tc_name', sa.String(length=256), nullable=False),
                 sa.Column('freq_list', sa.String(), nullable=False),
                 sa.Column('freq_nr', sa.INTEGER(), nullable=False),
                 sa.Column('insert_time', postgresql.TIMESTAMP(),
                           nullable=True),
                 sa.PrimaryKeyConstraint('id'),
                 sa.UniqueConstraint('utc_time', 'tc_name'),
                 schema='pipeline'
                 )
    create_table('sbm_log',
                 sa.Column('id', sa.BIGINT(), nullable=False),
                 sa.Column('sbm_type', sa.SMALLINT(), nullable=False),
                 sa.Column('cuc_time', sa.String(length=1024), nullable=False),
                 sa.Column('obt_time', postgresql.TIMESTAMP(), nullable=False),
                 sa.Column('utc_time', postgresql.TIMESTAMP(), nullable=True),
                 sa.Column('utc_time_is_predictive',
                           sa.BOOLEAN(), nullable=True),
                 sa.Column('sbm_qf', sa.FLOAT(), nullable=False),
                 sa.Column('sbm_algo', sa.SMALLINT(), nullable=False),
                 sa.Column('sbm_algo_param', postgresql.JSONB(
                     astext_type=sa.Text()), nullable=True),
                 sa.Column('insert_time', postgresql.TIMESTAMP(),
                           nullable=True),
                 sa.Column('selected', sa.BOOLEAN(), nullable=True),
                 sa.Column('status',
                           postgresql.ENUM(
                               'Available', 'Requested', 'Downlinked', 'Deleted', 'Missed', 'Unknown',
                               name='sbm_status_type'),
                           nullable=False),
                 sa.PrimaryKeyConstraint('id'),
                 sa.UniqueConstraint('cuc_time', 'sbm_type'),
                 schema='pipeline'
                 )
    create_table('event_log',
                 sa.Column('id', sa.BIGINT(), nullable=False),
                 sa.Column('start_time', postgresql.TIMESTAMP(),
                           nullable=False),
                 sa.Column('end_time', postgresql.TIMESTAMP(), nullable=False),
                 sa.Column('description', postgresql.JSONB(
                     astext_type=sa.Text()), nullable=False),
                 sa.Column('label', sa.String(), nullable=False),
                 sa.Column('is_predictive', sa.BOOLEAN(), nullable=True),
                 sa.Column('insert_time', postgresql.TIMESTAMP(),
                           nullable=True),
                 sa.PrimaryKeyConstraint('id'),
                 sa.UniqueConstraint('start_time', 'end_time',
                                     'label', 'description'),
                 schema='pipeline'
                 )
    # ### end Alembic commands ###

    # added by hand
    # fix table permissions
    grant_user_seq = (
        """GRANT USAGE, SELECT
        ON ALL SEQUENCES IN SCHEMA pipeline TO {0}""".format(user)
    )
    execute(grant_user_seq)


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    drop_table('sbm_log', schema='pipeline')
    drop_table('hfr_freq_list', schema='pipeline')
    drop_table('bia_sweep_table', schema='pipeline')
    drop_table('bia_sweep_log', schema='pipeline')
    drop_table('event_log', schema='pipeline')
    # ### end Alembic commands ###

    # delete the types if they exist
    execute('DROP TYPE IF EXISTS sbm_status_type')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Contains dingo tasks to synchronize the file tree with the ROC database."""

import os

from sqlalchemy import and_, or_
from sqlalchemy.dialects import postgresql
from sqlalchemy.orm import load_only

from poppy.core.logger import logger
from poppy.core.db.connector import Connector
from poppy.core.task import Task
from poppy.core.target import BaseTarget

from roc.dingo.constants import \
    PIPELINE_DATABASE, \
    ROC_DATA_ROOT, \
    FILE_LEVEL_LIST
from roc.dingo.models.file import FileLog

__all__ = ['CheckFileInDb']


class CheckFileInDb(Task):
    """
    Parse ROC file tree and synchronize with the the ROC database
    """
    plugin_name = 'roc.dingo'
    name = 'check_file_in_db'
    files_to_check = []
    files_to_insert = []
    files_to_update = []

    def add_targets(self):
        logger.debug('CheckFileInDb() : add_targets')

        self.add_input(target_class=BaseTarget, many=True,
                       identifier='roc_data_files')

        self.add_input(target_class=BaseTarget, many=True,
                       identifier='roc_dir_list')

        self.add_output(target_class=BaseTarget, many=True,
                        identifier='roc_data_files_to_insert')

        self.add_output(target_class=BaseTarget, many=True,
                        identifier='roc_data_files_to_update')

    def setup_inputs(self):
        """
        Setup task inputs.

        :param task:
        :return:
        """

        logger.debug('CheckFileInDb() : setup_inputs')

        # get files to check
        self.files_to_check = self.inputs['roc_data_files'].data

        # get the directory list
        self.dir_list = self.inputs['roc_dir_list'].data

        # get the root file tree
        self.root = self.pipeline.get('root', default=ROC_DATA_ROOT, args=True)
        # ensure that there is / at the end
        self.root = os.path.join(self.root, '')

        # get the directories to include
        self.level_list = self.pipeline.get('level',
                                            default=FILE_LEVEL_LIST,
                                            args=True)
        # get the force flag
        self.force = self.pipeline.get('force', default=False, args=True)

    def run(self):

        logger.debug('CheckFileInDb() : run')

        # Get the database connection if needed
        if not hasattr(self, 'session'):
            self.session = Connector.manager[PIPELINE_DATABASE].session

        # Initialize inputs
        self.setup_inputs()

        # Initialize outputs
        self.files_to_insert = self.files_to_check
        self.files_to_update = {}

        # Define filters to check file_log entry existence
        level_filter = FileLog.level.in_(self.level_list)
        dirname_filter = [
            FileLog.dirname.like(
                '%' + d.replace(self.root, '').rstrip(os.sep) + '%')
            for d in self.dir_list
        ]

        # Define relevant fields to retreive
        fields = ['basename', 'dirname', 'creation_time', 'size', 'is_removed']

        try:
            query = self.session.query(FileLog).\
                options(load_only(*fields)).\
                filter(and_(level_filter, or_(*dirname_filter)))
            logger.debug(
                str(query.statement.compile(
                    dialect=postgresql.dialect(),
                    compile_kwargs={'literal_binds': True})))
            results = query.all()
        except Exception:
            logger.error('Query failed :')
            logger.error(
                str(query.statement.compile(
                    dialect=postgresql.dialect(),
                    compile_kwargs={'literal_binds': True})))
        else:
            logger.debug('ok')
            for item in results:
                # find the corresponding file in the entry collection
                try:
                    item_file = self.files_to_insert.pop(item.basename)
                except Exception:
                    if not item.is_removed:
                        item.is_removed = True
                        item.state = 'OK'
                        item.error_log = None
                        self.session.commit()
                        logger.info(
                            'File {} in database not in root -> '
                            'mark as removed'.format(item.basename))
                    # else : the file was was already marked as removed
                    continue

                # Get item parameters
                creation_time = item_file['creation_time']
                size = item_file['size']
                dirname = os.path.dirname(
                    item_file['filepath']).replace(self.root, '')

                # dirname or size or creation_time is different => update
                if self.force:
                    logger.debug(
                        '{} to be updated [force]'.format(item.basename))
                    self.files_to_update[item.basename] = item_file
                elif (item.creation_time != creation_time) | \
                     (item.size != size) | \
                     (item.dirname != dirname):
                    self.files_to_update[item.basename] = item_file
                    logger.debug(
                        '{} already in database - has to be updated :'
                        '\n\tDirname : {} (db) / {} (fs)'
                        '\n\tSize : {} (db) / {} (fs)'
                        '\n\tDate : {} (db) / {} (fs)'.format(
                            item.basename,
                            item.dirname, dirname,
                            str(item.size), str(size),
                            str(item.creation_time), str(creation_time)))
                else:
                    logger.debug(
                        '{} already in database'.format(item.basename))
                    if item.is_removed:
                        # for any reason, file was marked as removed
                        logger.debug('-> is_removed set to False')
                        item.is_removed = False
                        self.session.commit()

        # Returns outputs
        self.outputs['roc_data_files_to_insert'].data = self.files_to_insert
        logger.info(
            '{} files to insert'.format(str(len(self.files_to_insert))))
        if len(self.files_to_insert) == 0:
            self.outputs['roc_data_files_to_insert'].is_empty = True

        self.outputs['roc_data_files_to_update'].data = self.files_to_update
        logger.info(
            '{} files to update'.format(str(len(self.files_to_update))))
        if len(self.files_to_update) == 0:
            self.outputs['roc_data_files_to_update'].is_empty = True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Contains dingo tasks to insert RPW L0 data into the ROC database."""
from datetime import datetime, timedelta
import json
import concurrent.futures

import pandas as pd
import scipy.sparse as sparse
import h5py
import uuid
from sqlalchemy.exc import IntegrityError
from sqlalchemy import and_, null

from poppy.core.logger import logger
from poppy.core.db.connector import Connector
from poppy.core.task import Task
from poppy.core.target import FileTarget

from roc.dingo.models.data import SbmLog, LfrKcoeffDump, BiaSweepLog, HfrTimeLog
from roc.dingo.models.packet import TmLog, TcLog
from roc.dingo.tools import get_packet_sha, compute_apid, load_spice, \
    query_db, is_sclk_uptodate, get_columns, get_current_idb, \
    bulk_insert, glob_paths, valid_time
from roc.dingo.constants import PIPELINE_DATABASE, \
    TIME_DAILY_STRFORMAT, TRYOUTS, TIME_WAIT_SEC, SQL_LIMIT, PIPELINE_TABLES, \
    SBM_ALGO_PARAM_LIST, LFR_KCOEFF_PARAM_NR, \
    SBM_LOG_PACKETS, LFR_KCOEFF_DUMP_PACKETS, \
    TIME_ISO_STRFORMAT, PACKET_TYPE, TC_ACK_ALLOWED_STATUS, IDB_SOURCE, WORKERS, \
    BIA_SWEEP_LOG_PACKETS, CIWT0130TM, \
    NAIF_SOLO_ID, PACKET_DATA_GROUP, EVENT_TM_APID, TC_ACK_APID, HK_TM_APID, HFR_SCIENCE_PACKETS

__all__ = ["L0ToDb", "get_l0_files", "load_param"]


class L0ToDb(Task):
    """
    Insert content of input L0 files into the
    in ROC database
    """
    plugin_name = 'roc.dingo'
    name = 'l0_to_db'

    def add_targets(self):
        self.add_input(identifier='rpw_l0_files',
                       many=True,
                       filepath=get_l0_files,
                       target_class=FileTarget)

    @Connector.if_connected(PIPELINE_DATABASE)
    def setup_inputs(self):

        # get the input l0 files
        self.l0_files = glob_paths(self.inputs['rpw_l0_files'].filepath)

        # Get start_time input value
        self.start_time = valid_time(self.pipeline.get(
            'start_time', default=[None])[0],
            format=TIME_DAILY_STRFORMAT)

        # Get end_time input value
        self.end_time = valid_time(self.pipeline.get(
            'end_time', default=[None])[0],
            format=TIME_DAILY_STRFORMAT)

        # Get include/exclude optional inputs
        self.include = self.pipeline.get(
            'include', default=[], create=True)
        self.exclude = self.pipeline.get(
            'exclude', default=[], create=True)

        # Retrieve --exclude-tm-apid, --exclude-tc-apid
        self.exclude_tm_apid = self.pipeline.get('exclude_tm_apid',
                                                 default=[],
                                                 args=True)
        self.exclude_tc_apid = self.pipeline.get('exclude_tc_apid',
                                                 default=[],
                                                 args=True)

        # Get or create failed_files list from pipeline properties
        self.failed_files = self.pipeline.get(
            'failed_files', default=[], create=True)

        # Get or create processed_files list from pipeline properties
        self.processed_files = self.pipeline.get(
            'processed_files', default=[], create=True)

        # get a database session
        self.session = Connector.manager[PIPELINE_DATABASE].session

        # Get tryouts from pipeline properties
        self.tryouts = self.pipeline.get(
            'tryouts', default=[TRYOUTS], create=True)[0]

        # Get wait from pipeline properties
        self.wait = self.pipeline.get(
            'wait', default=[TIME_WAIT_SEC], create=True)[0]

        # Get number of workers to run in parallel
        self.workers = self.pipeline.get(
            'workers', default=[WORKERS], create=True)[0]

        # Get idb_version/idb_source from pipeline properties
        self.idb_source = self.pipeline.get(
            'idb_source',
            default=[IDB_SOURCE],
            create=True)[0]

        self.idb_version = self.pipeline.get(
            'idb_version',
            default=[None],
            create=True)[0]

        # If idb_version not passed (is None),
        # then try to get current working version from the database
        if self.idb_version is None:
            self.idb_version = get_current_idb(
                self.idb_source,
                self.session,
                tryouts=self.tryouts,
                wait=self.wait,
            )
        if self.idb_version is None:
            raise ValueError(f'idb_version argument cannot be defined!')

        # Get SOLO SPICE kernels (SCLK and LSK)
        self.sclk_file = self.pipeline.get(
            'sclk', default=[], create=True)[0]
        self.lsk_file = self.pipeline.get(
            'lsk', default=[None], create=True)[0]
        if not self.sclk_file or not self.lsk_file:
            raise FileNotFoundError(
                'Both sclk_file and lsk_file must be passed as inputs to run L0ToDb!')
        else:
            # Load SPICE with input kernels
            self.spice = load_spice(
                spice_kernels=[self.lsk_file,
                               self.sclk_file]
            )

        # Retrieve --limit keyword value
        self.limit = self.pipeline.get('limit',
                                       default=[SQL_LIMIT],
                                       )[0]

        # Get --param-only optional keyword
        self.param_only = self.pipeline.get('param_only',
                                       default=False,
                                        )
        if self.param_only:
            logger.warning(f'Disable data insertion in tm_log/tc_log table')

        # Initialize data insertion counters
        self.inserted_count = 0

    def run(self):
        # Define task job ID (long and short)
        self.job_uuid = str(uuid.uuid4())
        self.job_id = f'L0ToDb-{self.job_uuid[:8]}'
        logger.info(f'Task {self.job_id} is starting')
        try:
            self.setup_inputs()
        except:
            logger.exception(
                f'Initializing inputs has failed for {self.job_id}!')
            self.pipeline.exit()
            return

        n_l0_files = len(self.l0_files)
        logger.info(f'{n_l0_files} RPW L0 files to process')
        if n_l0_files == 0:
            return

        # Loop over each L0 file in the input list
        for i, current_l0_file in enumerate(self.l0_files):
            logger.info(f'Processing {current_l0_file}    ({n_l0_files - i - 1} remaining)')
            self.current_l0_file = current_l0_file
            self.insert_time = datetime.today()
            # Open file
            try:
                with h5py.File(current_l0_file, 'r') as l0:
                    # Check if the SCLK SPICE kernel used to compute TM utc times in the input L0 file
                    # is older or newer than the l0 file date (if kernel is older, then it means that utc times are computed
                    # with predictive time coefficients. In this case, the utc times are not stored in the database
                    # in order to avoid confusion)
                    current_l0_datetime = datetime.strptime(l0.attrs['Datetime'][:8],
                                                            TIME_DAILY_STRFORMAT,
                                                            )
                    # Skip file if outside of the [start_time, end_time] time range
                    # (if any)
                    if self.start_time and self.start_time.date() > current_l0_datetime.date():
                        logger.info(f'{current_l0_file} older than {self.start_time.date()}, skip it')
                        continue
                    if self.end_time and self.end_time.date() < current_l0_datetime.date():
                        logger.info(f'{current_l0_file} newer than {self.end_time.date()}, skip it')
                        continue
                    # Retrieving L0 start/end times
                    self.l0_start_time = datetime.strptime(
                        l0.attrs['TIME_MIN'], TIME_ISO_STRFORMAT)
                    self.l0_end_time = datetime.strptime(
                        l0.attrs['TIME_MAX'], TIME_ISO_STRFORMAT)
                    try:
                        self.is_predictive_time = not is_sclk_uptodate(
                            current_l0_datetime + timedelta(days=1), l0.attrs['SPICE_KERNELS'])
                    except:
                        self.is_predictive_time = True

                        if self.is_predictive_time:
                            logger.info(f'Predictive UTC times used in {current_l0_file}!')

                    # Initialize existing packet Dataframe
                    self.existing_data = pd.DataFrame()

                    # Get existing packet SHA in database between
                    # l0 start/end times
                    query_start = self.l0_start_time - timedelta(minutes=1)
                    query_end = self.l0_end_time + timedelta(minutes=1)
                    logger.debug(f'Getting existing TmLog.sha between {query_start} and {query_end}')
                    existing_tm = self._get_existing_data(TmLog, TmLog.sha,
                                                            start_time=query_start,
                                                            end_time=query_end)
                    logger.info(f'{existing_tm.shape[0]} TmLog.sha entries found between {query_start} and {query_end}')
                    if existing_tm.shape[0] > 0:
                        self.existing_data = pd.concat(
                            [self.existing_data, existing_tm]).reset_index(drop=True)
                    logger.debug(f'Getting existing TcLog.sha between {query_start} and {query_end}')
                    existing_tc = self._get_existing_data(TcLog, TcLog.sha,
                                                        start_time=query_start,
                                                        end_time=query_end)
                    logger.info(
                        f'{existing_tc.shape[0]} TmLog.sha entries found between {query_start} and {query_end}')
                    if existing_tc.shape[0] > 0:
                        self.existing_data = pd.concat([self.existing_data,
                            existing_tc]).reset_index(drop=True)

                    # Insert input L0 data into database
                    try:
                        inserted_count = self.insert_l0(l0,
                                                        include=self.include,
                                                        exclude=self.exclude)
                    except:
                        self.failed_files.append(self.current_l0_file)
                        logger.exception(f'Insertion has failed for {self.current_l0_file} in task {self.job_id}')
                        break
                    else:
                        logger.info(f'{inserted_count} new packets '
                                    f'inserted for {self.current_l0_file}')
                        self.inserted_count += inserted_count
                        self.processed_files.append(self.current_l0_file)
            except:
                self.failed_files.append(self.current_l0_file)
                logger.exception(f'Cannot open {self.current_l0_file} in task {self.job_id}!')
                break

        logger.info(f'{self.inserted_count} new packets inserted in the database'
                    f' from files: {", ".join(self.processed_files)}')
        if len(self.failed_files) > 0:
            logger.error(f'Insertion has failed for files: {", ".join(self.failed_files)} in task {self.job_id}')

        logger.info(f'Task {self.job_id} has ended correctly')

    def insert_l0(self, l0,
                  packet_type=PACKET_TYPE,
                  include=[],
                  exclude=[]):
        """
        Insert L0 packet data into the database.
        Multiple packet insertions in parallel is possible.

        :param l0: h5.group object containing current L0 packet data
        :param packet_type: Filter by type of packets (TM, TC)
        :param include: List of packets to insert (all by default)
        :param exclude: List of exclude from insertion (empty list by default)
        :return: Number of inserted packets
        """
        inserted_count = 0
        # Build list of packets to insert
        # (Filter if required using include/exclude keywords)
        packet_list = []
        packet_count = 0
        for current_type in packet_type:
            packet_count += int(l0[current_type].attrs['COUNT'])
            packet_list.extend([f'{current_type}/{current_packet}'
                                for current_packet in l0[current_type].keys()
                                if self.is_packet_valid(current_packet,
                                                        include=include,
                                                        exclude=exclude)])

        logger.info(f'{packet_count} packets from {len(l0["TM"].keys())} TM and {len(l0["TC"].keys())} TC found in {self.current_l0_file}   (workers={self.workers})')

        # Loop over each TM/TC packet in L0
        # We can use a with statement to ensure threads are cleaned up promptly
        with concurrent.futures.ThreadPoolExecutor(max_workers=self.workers) as executor:
            # Start the load operations and mark each future with its file
            future_to_packet = {executor.submit(self.insert_packet,
                                                l0,
                                                current_packet,
                                                ):
                                current_packet for current_packet in packet_list}

            for i, future in enumerate(concurrent.futures.as_completed(future_to_packet)):
                current_packet = future_to_packet[future]
                try:
                    n_pkt_to_insert = future.result()
                except:
                    logger.exception(f'Bulk insertion has failed for {current_packet} in {self.current_l0_file} for task {self.job_id}!')
                else:
                    logger.info(f'{n_pkt_to_insert} {current_packet} new packets inserted from {self.current_l0_file}')
                    inserted_count += n_pkt_to_insert

        return inserted_count

    def insert_packet(self, l0, packet_name):
        """
        Insert data for the input packet name found in L0 file

        :param l0: h5.group (root) in the L0 file
        :param packet_name: Name of packet for which data must be inserted
        :return: number of packets inserted
        """

        # Get table class
        current_type = packet_name[:2]
        current_model = PIPELINE_TABLES[f'{current_type.lower()}_log']

        # Convert input l0 packet data to pandas.DataFrame
        logger.debug(f'Preparing {packet_name} data for insertion for {self.current_l0_file}')
        current_df = self.packet_to_dataframe(l0, packet_name)

        # Remove already inserted packet data
        # Only keep unique SHA
        if current_df.shape[0] > 0:
            if self.existing_data.shape[0] > 0:
                logger.debug(f'Removing {self.existing_data.shape[0]} {packet_name} data already inserted for {self.current_l0_file}')
                current_df = current_df[
                    ~current_df.sha.isin(self.existing_data.sha)]

        n_pkt_to_insert = current_df.shape[0]
        if n_pkt_to_insert == 0:
            logger.debug(f'There is no new {packet_name} to insert in {self.current_l0_file}')
        else:
            if not self.param_only:
                logger.debug(f'Inserting {n_pkt_to_insert} {packet_name} from {self.current_l0_file}')
                # Only keep tm_log/tc_log columns to insert
                if 'data' not in current_df.columns:
                    columns = get_columns(current_model, remove=[
                                          'id', 'binary', 'data'])
                else:
                    columns = get_columns(current_model, remove=['id', 'binary'])
                # convert to list of dictionaries
                data_to_insert = current_df[columns].to_dict('records')
                # Insert bulk of data
                bulk_insert(self.session, current_model, data_to_insert)
            else:
                logger.debug(f'Inserting {packet_name} packet parameters from {self.current_l0_file}')

            # Insert other tables data
            self._insert_packet_param(packet_name, current_df)

        return n_pkt_to_insert

    def packet_to_dataframe(self, l0, packet_name):
        """
        Convert data of input l0 packet into pandas.DataFrame object.

        :param l0: h5.group containing l0 file packets data
        :param packet_name: name of the packet for which data must be converted
        :return: instance of pandas.DataFrame containing packet data
        """

        # Get packet type (TM or TC)
        packet_type = packet_name[:2]
        data_grp = PACKET_DATA_GROUP[packet_type]

        # Get number of packets
        packet_nsamp = l0[packet_name].attrs['COUNT']

        # Get packet header
        packet_header = load_param(l0[packet_name][
            'packet_header'])
        # Get packet data field header
        data_field_header = load_param(l0[packet_name][
            'data_field_header'])

        # Compute APID of the packet
        packet_apid = compute_apid(packet_header.process_id[0],
                                   packet_header.packet_category[0])

        # If exclude_tm/tc_apid not empty list,
        # then skip packet if its APID in the exclude_tm/tc_apid list
        if packet_type == 'TM' and self.exclude_tm_apid and packet_apid in self.exclude_tm_apid:
            logger.info(f'{packet_name} ({packet_apid}) is in excluded apid list ({self.exclude_tm_apid})')
            return pd.DataFrame()
        elif packet_type == 'TC' and self.exclude_tc_apid and packet_apid in self.exclude_tc_apid:
            logger.info(f'{packet_name} ({packet_apid}) is in excluded apid list ({self.exclude_tc_apid})')
            return pd.DataFrame()

        # Get packet data
        packet_data = load_param(l0[packet_name][
            data_grp])

        # Concatenate packet headers and data dataframes
        packet_df = pd.concat(
            [packet_header, data_field_header, packet_data], axis=1)

        # Use a dict first to store data (to avoid Pandas PerformanceWarning)
        packet_info = dict()

        # Add insertion time
        packet_info['insert_time'] = [self.insert_time] * packet_nsamp

        # Get packet_length
        packet_info['length'] = packet_df['packet_length']

        # Add APID
        packet_info['apid'] = [packet_apid] * packet_nsamp

        # Store packet utc time values
        packet_info['utc_time'] = l0[packet_name]['utc_time'][()]

        # Store packet binary data (as hexa string)
        packet_info['binary'] = l0[packet_name]['binary'][()]

        packet_info['utc_time_is_predictive'] = [
            self.is_predictive_time] * packet_nsamp

        packet_info['palisade_id'] = [
            packet_name.split('/')[-1]] * packet_nsamp
        packet_info['srdb_id'] = [l0[packet_name].attrs[
            'SRDB_ID']] * packet_nsamp
        packet_info['category'] = [l0[packet_name].attrs[
            'PACKET_CATEGORY']] * packet_nsamp

        # Add info to dataframe
        packet_df = pd.concat(
            [packet_df, pd.DataFrame.from_dict(packet_info)],
            axis=1)

        # Convert utc_time byte strings into datetime objects
        packet_df['utc_time'] = packet_df['utc_time'].apply(
            lambda x: datetime.strptime(x.decode('UTF-8')[: -4] + 'Z', TIME_ISO_STRFORMAT))

        # Convert binary from bytes to string
        packet_df['binary'] = packet_df['binary'].apply(
            lambda x: x.decode('UTF-8'))

        # Case of compressed TM packets
        if 'compressed' in l0[packet_name].keys():
            # Be sure that compressed TM have the right
            # PALISADE_ID, SDRB_ID and CATEGORY
            where_compressed = (l0[
                packet_name]['compressed'][()] == 1)
            packet_df.loc[where_compressed, 'srdb_id'] = l0[packet_name].attrs[
                'SRDB_ID_C']
            packet_df.loc[where_compressed, 'category'] = l0[packet_name].attrs[
                'PACKET_CATEGORY_C']
            packet_df.loc[where_compressed,
                          'palisade_id'] = packet_name + '_C'

        if packet_type == 'TM':

            # Init dict (to avoid Pandas PerformanceWarning)
            packet_info = dict()

            # Get cuc_time
            packet_info['cuc_time'] = packet_df['time'].apply(
                lambda x: self.spice.cuc2str(x[0], x[1]))

            # Get obt_time
            packet_info['obt_time'] = packet_df['time'].apply(
                lambda x: self.spice.cuc2datetime(x[0], x[1]))

            # Get packet sync_flag
            packet_info['sync_flag'] = packet_df['time'].apply(
                lambda x: x[2] == 0)

            # Get TM source data
            packet_info['data'] = [null()] * packet_nsamp

            # Add to packet_df dataframe
            packet_df = pd.concat(
                [packet_df, pd.DataFrame.from_dict(packet_info)],
                axis=1)

            try:
                # Only store event and TC ack. TM packets in the
                # database
                apid_to_store = EVENT_TM_APID + TC_ACK_APID + HK_TM_APID
                where_apid = (packet_df['apid'].isin(apid_to_store))
                packet_df.loc[where_apid, ('data')] = packet_df.loc[
                    where_apid, l0[packet_name][data_grp].keys()].to_dict('records')
            except:
                # If failed, assume there is no parameter for this TM
                pass

        elif packet_type == 'TC':

            # Init dict (to avoid Pandas PerformanceWarning)
            packet_info = dict()

            packet_info['unique_id'] = l0[
                packet_name]['unique_id'][()].astype(str)
            packet_info['sequence_name'] = l0[
                packet_name]['sequence_name'][()].astype(str)
            current_tc_state = l0[
                packet_name]['tc_ack_state'][()].astype(str)
            packet_info['tc_acc_state'] = current_tc_state[:, 0]
            packet_info['tc_exe_state'] = current_tc_state[:, 1]

            # Get TC application data
            try:
                packet_info['data'] = packet_df[
                    l0[packet_name][data_grp].keys()].to_dict('records')
            except:
                # If failed, assume there is no parameter for this TC
                pass

            # Add to packet_df dataframe
            packet_df = pd.concat(
                [packet_df, pd.DataFrame.from_dict(packet_info)],
                axis=1)

            # Only store PASSED/FAILED TC in database
            packet_df = packet_df.loc[packet_df[
                'tc_exe_state'].isin(TC_ACK_ALLOWED_STATUS)]

        # If there is no packet anymore to insert, then return empty DataFrame
        if packet_df.shape[0] == 0:
            logger.debug(f'{packet_name} DataFrame is empty in {self.current_l0_file}!')
            return pd.DataFrame()

        # Compute SHA
        packet_info = {'sha': packet_df.apply(
            lambda x: get_packet_sha(x),
            axis=1)}
        # Add SHA to dataframe
        packet_df = pd.concat([packet_df, pd.DataFrame.from_dict(packet_info)],
                              axis=1)

        # Make sure to have unique packets (unique SHA values)
        packet_df.drop_duplicates(subset=['sha'], inplace=True)

        return packet_df

    def is_packet_valid(self, packet_name,
                        include=[],
                        exclude=[]):
        """
        Check if input packet name is
        in the include and/or exclude lists

        :param packet_name:
        :param include: list of packets to include (if empty list, then return True)
        :param exclude: list of packets to exclude (default is empty list)
        :return: True if packet_name is include and/or not in exclude
        """
        is_valid = True
        if packet_name in exclude:
            logger.debug(f'{packet_name} excluded')
            is_valid = False
        if include and packet_name not in include:
            logger.debug(f'{packet_name} not in {include}')
            is_valid = False
        return is_valid

    def _insert_packet_param(self, packet_name, packet_data):
        """
        Insert following data for current packet:
             - sbm_log
             - bia_sweep_log
             - lfr_kcoeff_dump

        It is assumed that input data are not already stored in the
        database.

        :param packet_name: Name of the input packet
        :param packet_data: packet data as a pandas.DataFrame
        :return: number of packets inserted
        """
        n_packet = packet_data.shape[0]
        packet_name = packet_name.split('/')[-1]
        # Set inputs corresponding to input packet
        if packet_name in SBM_LOG_PACKETS:
            logger.info(f'Inserting {packet_name} data into pipeline.sbm_log table')
            model = SbmLog
            columns = get_columns(model, remove=['id', 'retrieved_time'])
            sbm_type = int(packet_name[-1])
            packet_data['sbm_type'] = [sbm_type] * n_packet
            packet_data['selected'] = [False] * n_packet
            packet_data['status'] = ["Unknown"] * n_packet
            packet_data['sbm_qf'] = packet_data[f'HK_RPW_S20_SBM{sbm_type}_QF_D'].astype(float)
            packet_data['sbm_algo'] = packet_data[f'SY_DPU_SBM{sbm_type}_ALGO'].astype(int)
            packet_data['sbm_algo_param'] = packet_data.apply(
                lambda x: self.extract_sbm_algo_param(x),
                axis=1)

            # Compute event occurred time
            occurred_time_key = f'HK_RPW_S20_SBM{sbm_type}_TIME_D'
            packet_data['cuc_time'] = packet_data[occurred_time_key].apply(
                lambda x: self.spice.cuc2str(x[0], x[1]))
            packet_data['obt_time'] = packet_data[occurred_time_key].apply(
                lambda x: self.spice.cuc2datetime(x[0], x[1]))
            packet_data['utc_time'] = packet_data[occurred_time_key].apply(
                self.cuc2utc)

        elif packet_name in BIA_SWEEP_LOG_PACKETS:
            logger.info(f'Inserting {packet_name} data into pipeline.bia_sweep_log table')
            model = BiaSweepLog
            columns = get_columns(model, remove=['id'])
            packet_data['sweep_step'] = packet_data.apply(
                lambda x: CIWT0130TM[int(x['PA_DPU_BIA_SWEEP_PR_CODE'])],
                axis=1,
            )
            packet_data['utc_time'] = packet_data[
                'PA_DPU_BIA_SWEEP_TIME'].apply(self.cuc2utc)
        elif packet_name in LFR_KCOEFF_DUMP_PACKETS:
            logger.info(f'Inserting {packet_name} data into pipeline.lfr_kcoeff_dump table')
            model = LfrKcoeffDump
            columns = get_columns(model, remove=['id'])
            packet_data["kcoeff_pkt_cnt"] = packet_data[
                'PA_LFR_KCOEFF_PKT_CNT'].astype(int)
            packet_data["kcoeff_pkt_nr"] = packet_data[
                'PA_LFR_KCOEFF_PKT_NR'].astype(int)
            packet_data["kcoeff_blk_nr"] = packet_data[
                'PA_LFR_KCOEFF_BLK_NR'].astype(int)
            packet_data["kcoeff_values"] = packet_data.apply(
                lambda x: self.extract_lfr_kcoeff(x),
                axis=1
            )
        else:
            # Otherwise exit insertion normally
            # logger.debug(f'No extra data to insert for {packet_name}')
            return True

        data_to_insert = packet_data[columns].to_dict('records')
        try:
            bulk_insert(self.session, model, data_to_insert)
        except IntegrityError:
            logger.error(f'Bulk insertion in table {model} may be incompleted '
                         f'for {packet_name} in {self.current_l0_file} for task {self.job_id}')
            n_packet = 0
        except:
            logger.exception(f'Bulk insertion in table {model} has failed '
                             f'for {packet_name} in {self.current_l0_file} for task {self.job_id}')
            n_packet = 0

        return n_packet

    def cuc2utc(self, cuc_time, naif_id=NAIF_SOLO_ID):
        """
        Convert input RPW CUC time into UTC time

        :param cuc_time:
        :return: UTC time as returned by SpiceManager.obt2utc() method
        """
        obt_time = self.spice.cuc2obt(cuc_time)
        return self.spice.obt2utc(naif_id, obt_time)

    def _get_existing_data(self, model, fields,
                           start_time=None,
                           end_time=None,
                           to_dict=None):
        """
        Query database to return existing data for a given table

        :param model: class of the table
        :param fields: fields to query
        :param start_time: Filter query by start time
        :param end_time: filter query by end time
        :param to_dict: See pandas.DataFrame.to_dict()
        :return: returned rows (as a DataFrame by default)
        """

        # Get list of existing data in the database
        filters = list()
        # Add start_time/end_time filters (if passed)
        if start_time:
            filters.append(model.utc_time >= str(
                start_time - timedelta(hours=1)))
        if end_time:
            filters.append(model.utc_time <= str(
                end_time + timedelta(hours=1)))

        if fields is None:
            fields = model

        results = query_db(self.session, fields,
                           filters=(and_(*filters) if filters else None),
                           tryouts=self.tryouts,
                           wait=self.wait,
                           limit=self.limit,
                           to_dict=to_dict)

        return results

    def extract_sbm_algo_param(self, current_packet):
        """
        Extract SBM algo parameters from current packet

        :param current_packet: current packat data as a pandas.DataFrame
        :return: list of parameters returned as a JSON string
        """
        current_sbm_algo_param = {
            current_param: current_packet[current_param]
            for current_param in SBM_ALGO_PARAM_LIST[current_packet['sbm_type']]
        }
        # Store SBM algo parameters as a JSON string
        return current_sbm_algo_param

    def extract_lfr_kcoeff(self, packet_data):
        """
        Extract LFR Kcoeff parameters from current TM
        kcoeff data are returned in JSON format, where
        keyword is the frequency index and values are arrays of [LFR_KCOEFF_PARAM_NR, kcoeff_blk_nr] samples

        :param packet_data: current TM packet_data
        :return:json.dumps object
        """
        kcoeffs = {}
        blk_nr = packet_data['kcoeff_blk_nr']
        for current_freq in packet_data['SY_LFR_KCOEFF_FREQUENCY']:
            kcoeffs[current_freq] = ",".join([
                str(packet_data[f'SY_LFR_KCOEFF_{j + 1}'][0:blk_nr])
                for j in range(LFR_KCOEFF_PARAM_NR)])

        return json.dumps(kcoeffs)


def get_l0_files(pipeline):
    try:
        l0_files = pipeline.args.rpw_l0_files
        if not isinstance(l0_files, list):
            l0_files = [l0_files]
        return sorted(l0_files)
    except:
        # If not defined as input argument, then assume that it is already
        # defined as target input
        pass


def load_param(current_group):
    """
    Sub-methods to save h5py.Data array in
    a pandas.DataFrame

    :param current_group:
    :return: DataFrame with h5py.Data
    """
    # Input group must be a h5py.Group object
    if not isinstance(current_group, h5py.Group):
        return pd.DataFrame()

    current_df = dict()
    for key, val in current_group.items():
        current_val = val[()]
        if len(current_val.shape) > 1:
            arr = sparse.coo_matrix(current_val)
            current_df[key] = arr.toarray().tolist()
        else:
            current_df[key] = current_val

    return pd.DataFrame.from_dict(current_df)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Reference documentation: ROC-GEN-SYS-NTT-00038-LES

"""
Database model for:
    - sbm_log,
    - lfr_kcoeff_dump
    - bia_sweep_log,
    - efecs_events
    - event_log
    - solohk_param
    - process_queue
tables.
"""

from poppy.core.db.non_null_column import NonNullColumn

from poppy.core.db.base import Base
from sqlalchemy import String, UniqueConstraint
from sqlalchemy.dialects.postgresql import (
    BIGINT,
    BOOLEAN,
    FLOAT,
    ENUM,
    SMALLINT,
    TIMESTAMP,
    JSONB,
)

__all__ = [
    'SbmLog',
    'BiaSweepLog',
    'LfrKcoeffDump',
    'EventLog',
    'SoloHkParam',
    'ProcessQueue',
    'EfecsEvents',
    'HfrTimeLog',
    'Event'
]

# Create enumeration for sbm_log.status column
SBM_STATUS_LIST = ['Available', 'Requested',
                   'Downlinked', 'Deleted', 'Missed', 'Unknown']
sbm_status_enum = ENUM(
    *SBM_STATUS_LIST,
    name='sbm_status_type',
    # schema='pipeline',
)


class SbmLog(Base):
    """
    Class representation of the table for sbm_log table in the ROC database.
    """

    id = NonNullColumn(BIGINT(), primary_key=True)
    sbm_type = NonNullColumn(
        SMALLINT(),
        descr='Type of SBM event (1=SBM1 or 2=SBM2)')
    cuc_time = NonNullColumn(
        String(1024),
        descr='SBM event occurrence on-board time in CCSDS CUC'
              'format "coarse:fine".')
    obt_time = NonNullColumn(
        TIMESTAMP(),
        descr='SBM event occurrence on-board time in SQL timestamp format')
    utc_time = NonNullColumn(
        TIMESTAMP(),
        nullable=True,
        descr='SBM event occurrence UTC time in SQL timestamp format')
    utc_time_is_predictive = NonNullColumn(
        BOOLEAN,
        nullable=True,
        descr='Flag to indicate if UTC time is predictive (True)'
              'or definitive (False)')
    sbm_qf = NonNullColumn(
        FLOAT(),
        descr='SBM detection quality factor')
    sbm_algo = NonNullColumn(
        SMALLINT(),
        descr='SBM detection algorithm status')
    sbm_algo_param = NonNullColumn(
        JSONB(),
        nullable=True,
        descr='List of SBM algo parameters (JSON format)')
    retrieved_time = NonNullColumn(
        TIMESTAMP(),
        nullable=True,
        descr='Local date/time at which the SBM event science data'
              'has been retrieved.')
    selected = NonNullColumn(
        BOOLEAN,
        nullable=True,
        descr='True if SBM event is selected')
    status = NonNullColumn(
        sbm_status_enum,
        descr='Status of the sbm event.'
              f'Possible values are: {SBM_STATUS_LIST}')
    insert_time = NonNullColumn(
        TIMESTAMP(),
        nullable=True,
        descr='Database insertion local time.')
    __tablename__ = 'sbm_log'
    __table_args__ = (
        UniqueConstraint('cuc_time', 'sbm_type'),
        {
            'schema': 'pipeline',
        }
    )


class BiaSweepLog(Base):
    """
    Class representation of the table for bia_sweep_log table
    in the ROC database.
    """
    id = NonNullColumn(BIGINT(), primary_key=True)
    sweep_step = NonNullColumn(
        String(16),
        descr='Step of the Bias sweep'
              '(TM_DPU_EVENT_PR_BIA_SWEEP.PA_DPU_BIA_SWEEP_PR_CODE values)')
    utc_time = NonNullColumn(
        TIMESTAMP(),
        descr='Sweep step UTC time')
    cuc_time = NonNullColumn(
        String(1024),
        descr='Sweep step on-board time in CCSDS CUC format "coarse:fine".')
    utc_time_is_predictive = NonNullColumn(
        BOOLEAN, nullable=True,
        descr='Flag to indicate if UTC time is predictive (True)'
              'or definitive (False)')
    insert_time = NonNullColumn(
        TIMESTAMP(),
        nullable=True,
        descr='Database insertion local time.')

    __tablename__ = 'bia_sweep_log'
    __table_args__ = (
        UniqueConstraint('sweep_step', 'utc_time'),
        {
            'schema': 'pipeline',
        }
    )


class LfrKcoeffDump(Base):
    """
    Class representation of the table for lfr_kcoeff_dump table
    in the ROC database.
    """

    id = NonNullColumn(BIGINT(), primary_key=True)
    utc_time = NonNullColumn(
        TIMESTAMP(),
        descr='TM_LFR_KCOEFFICIENTS_DUMP packet creation time (UTC)')
    cuc_time = NonNullColumn(
        String(1024),
        descr='Packet creation on-board time in CCSDS CUC'
              'format "coarse:fine".')
    utc_time_is_predictive = NonNullColumn(
        BOOLEAN,
        nullable=True,
        descr='Flag to indicate if UTC time is predictive (True)'
              'or definitive (False)')
    kcoeff_pkt_cnt = NonNullColumn(
        SMALLINT(),
        descr='Total count of packets for LFR inter calibration factors dump.'
              '(PA_LFR_KCOEFF_PKT_CNT)')
    kcoeff_pkt_nr = NonNullColumn(
        SMALLINT(),
        descr='Number of the packet for LFR inter calibration factors dump.'
              '(PA_LFR_KCOEFF_PKT_NR)')
    kcoeff_blk_nr = NonNullColumn(
        SMALLINT(),
        descr='Number of block LFR_KCOEFFICENT_PARAMETERS in the packet'
              '(PA_LFR_KCOEFF_BLK_NR)')
    kcoeff_values = NonNullColumn(
        JSONB(),
        descr='32 values of the Kcoeff for the current list of frequencies'
              '(json format)')
    insert_time = NonNullColumn(
        TIMESTAMP(),
        nullable=True,
        descr='Database insertion local time.')

    __tablename__ = 'lfr_kcoeff_dump'
    __table_args__ = (
        UniqueConstraint('utc_time', 'kcoeff_pkt_nr'),
        {
            'schema': 'pipeline',
        }
    )


class EventLog(Base):
    """
    Class representation of the table for event_log table in the ROC database.
    """
    id = NonNullColumn(BIGINT(), primary_key=True)
    start_time = NonNullColumn(
        TIMESTAMP(),
        descr='Event UTC start time')
    end_time = NonNullColumn(
        TIMESTAMP(),
        descr='Event UTC end time')
    description = NonNullColumn(
        JSONB(),
        descr='Description of event')
    label = NonNullColumn(
        String(),
        descr='Label of event')
    is_predictive = NonNullColumn(
        BOOLEAN,
        nullable=True,
        descr='Flag to indicate if event is predictive (True)'
              'or definitive (False)')
    insert_time = NonNullColumn(
        TIMESTAMP(),
        nullable=True,
        descr='Database insertion local time.')

    __tablename__ = 'event_log'
    __table_args__ = (
        UniqueConstraint('start_time', 'end_time', 'label', 'description'),
        {
            'schema': 'pipeline',
        }
    )


class SoloHkParam(Base):
    """
    Class representation of the table solo_hk_param in the ROC database.
    """

    id = NonNullColumn(BIGINT(), primary_key=True)
    name = NonNullColumn(
        String(16),
        descr='Parameter name (SRDB ID)')
    utc_time = NonNullColumn(
        TIMESTAMP(),
        descr='ParamSampleListElement.TimeStampAsciiA XML tag value')
    description = NonNullColumn(
        String(),
        descr='ParamSampleListElement.Description XML tag value')
    unit = NonNullColumn(
        String(),
        descr='ParamSampleListElement.Unit XML tag value')
    eng_value = NonNullColumn(
        String(),
        descr='ParamSampleListElement.EngineeringValue XML tag value')
    raw_value = NonNullColumn(
        String(),
        descr='ParamSampleListElement.RawValue XML tag value')
    sha = NonNullColumn(
        String(),
        descr='SHA of the element. Computed from'
              '(name, time_stamp, raw_value)')

    __tablename__ = 'solo_hk_param'
    __table_args__ = (
        UniqueConstraint('sha'),
        {
            'schema': 'pipeline',
        }
    )


class ProcessQueue(Base):
    """
    Class representation of the table process_queue in the ROC database.
    """

    id = NonNullColumn(BIGINT(), primary_key=True)
    dataset_id = NonNullColumn(
        String(512),
        descr='Dataset ID in the ROC database')
    start_time = NonNullColumn(
        TIMESTAMP(),
        descr='Start time of data')
    end_time = NonNullColumn(
        TIMESTAMP(),
        nullable=True,
        descr='End time of data')
    version = NonNullColumn(
        String(16),
        nullable=True,
        descr='Data version of the file')
    file = NonNullColumn(
        String(),
        nullable=True,
        descr='Dataset file name'
    )
    insert_time = NonNullColumn(
        TIMESTAMP(),
        nullable=True,
        descr='Database insertion local time.')

    __tablename__ = 'process_queue'
    __table_args__ = (
        {
            'schema': 'pipeline',
        }
    )


class EfecsEvents(Base):
    """
    Class representation of the table efecs_events in the ROC database.
    """

    id = NonNullColumn(BIGINT(), primary_key=True)
    name = NonNullColumn(
        String(),
        descr='EFECS event name')
    utc_time = NonNullColumn(
        TIMESTAMP(),
        descr='EFECS event time')
    attributes = NonNullColumn(
        JSONB(),
        descr='EFECS event attributes (in JSON format)',
        nullable=True)
    ltp_count = NonNullColumn(
        SMALLINT(),
        descr='LTP counter')
    gen_time = NonNullColumn(
        TIMESTAMP(),
        descr='EFECS generation time')

    __tablename__ = 'efecs_events'
    __table_args__ = (
        UniqueConstraint('name', 'utc_time'),
        {
            'schema': 'pipeline',
        }
    )


class HfrTimeLog(Base):
    """
    Class representation of the table hfr_time_log in the ROC database.
    See https://gitlab.obspm.fr/ROC/RCS/THR_CALBAR/-/issues/76 for details
    about why this table is needed
    """

    id = NonNullColumn(BIGINT(), primary_key=True)
    acq_time = NonNullColumn(
        TIMESTAMP(),
        descr='HFR TM PA_THR_ACQUISITION_TIME human readable time',
    )
    coarse_time = NonNullColumn(
        BIGINT(),
        descr='HFR TM PA_THR_ACQUISITION_TIME, coarse part')
    fine_time = NonNullColumn(
        BIGINT(),
        descr='HFR TM PA_THR_ACQUISITION_TIME, fine part')
    mode = NonNullColumn(
        SMALLINT(),
        descr='HFR mode (0=NORMAL, 1=BURST)',
    )
    delta_time1 = NonNullColumn(
        JSONB(),
        descr='Values of delta_time for HF1 for each TM '
              'for the current PA_THR_ACQUISITION_TIME value. '
              'Keywords are packet creation times (coarse * 100000 + fine)',
        nullable=True
    )
    delta_time2 = NonNullColumn(
        JSONB(),
        descr='Values of delta_time for HF2 for each TM '
              'for the current PA_THR_ACQUISITION_TIME value. '
              'Keywords are packet creation times (coarse * 100000 + fine)',
        nullable=True
    )

    __tablename__ = 'hfr_time_log'
    __table_args__ = (
        UniqueConstraint('coarse_time', 'fine_time', 'mode'),
        {
            'schema': 'pipeline',
        }
    )


# Create enumeration for event.origin column
EVENT_ORIGIN_LIST = ['SOLO', 'RPW']
event_origin_enum = ENUM(
    *EVENT_ORIGIN_LIST,
    name='event_origin_enum',
    # schema='pipeline',
)


class Event(Base):
    """
    Class listing the different events for RPW
    """

    id = NonNullColumn(BIGINT(), primary_key=True)
    label = NonNullColumn(
        String(),
        descr='Event label'
    )
    is_tracked = NonNullColumn(
        BOOLEAN, default=True,
        descr='True if event has to be tracked'
    )
    is_anomaly = NonNullColumn(
        BOOLEAN, default=True,
        descr='True if event is an anomaly'
    )
    origin = NonNullColumn(
        event_origin_enum,
        descr='Origin of the event.'
              f'Possible values are: {EVENT_ORIGIN_LIST}'
    )

    __tablename__ = 'events'
    __table_args__ = (
        UniqueConstraint('label'),
        {
            'schema': 'pipeline',
        }
    )



# -*- coding: utf-8 -*-
"""Fix database discrepancy with model

Revision ID: roc_dingo_0010b_fix_db
Revises: roc_dingo_0010_add_view
Create Date: 2021-07-20 13:56:44.889371

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = 'roc_dingo_0013_fix_db'
down_revision = 'roc_dingo_0012_add_data'
branch_labels = None
depends_on = None


def upgrade():

    op.execute('COMMIT')
    # state columns type should be of type file_state_list
    # (and not file_state_type)
    op.execute('DROP TYPE IF EXISTS file_state_list')
    op.execute('ALTER TYPE file_state_type RENAME TO file_state_list')

    # idem for status
    op.execute('DROP TYPE IF EXISTS file_status_list')
    op.execute('ALTER TYPE file_status_type RENAME TO file_status_list')


def downgrade():
    op.execute('COMMIT')
    op.execute('ALTER TYPE file_state_list RENAME TO file_state_type')
    op.execute('ALTER TYPE file_status_list RENAME TO file_status_type')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Contains dingo tasks to update utc times (computed with SPICE kernels)
into the ROC database."""

from datetime import datetime, timedelta
from pathlib import Path
import uuid

from sqlalchemy import and_

from poppy.core.logger import logger
from poppy.core.db.connector import Connector
from poppy.core.task import Task

from roc.dingo.constants import PIPELINE_DATABASE, TRYOUTS, TIME_WAIT_SEC, SQL_LIMIT, NAIF_SOLO_ID
from roc.dingo.models.data import SbmLog, BiaSweepLog, LfrKcoeffDump
from roc.dingo.models.packet import TmLog
from roc.dingo.tools import load_spice, query_db, is_sclk_uptodate, valid_time


class UpdateUtcTime(Task):
    """
    Update UTC times in the
    in ROC database
    """
    plugin_name = 'roc.dingo'
    name = 'update_utc_time'

    def add_targets(self):
        pass

    @Connector.if_connected(PIPELINE_DATABASE)
    def setup_inputs(self):

        # Get SOLO SPICE kernels (SCLK and LSK)
        self.sclk_file = self.pipeline.get(
            'sclk', default=[None], create=True)[0]
        self.lsk_file = self.pipeline.get(
            'lsk', default=[None], create=True)[0]
        if not self.sclk_file or not self.lsk_file:
            raise FileNotFoundError(
                'Both sclk_file and lsk_file must be passed as inputs to run UpdateUtcTime!')
        else:
            # Load SPICE kernels
            self.spice = load_spice(spice_kernels=[self.lsk_file,
                                                   self.sclk_file])

        # Get start_time input value
        self.start_time = valid_time(self.pipeline.get(
            'start_time', default=[None])[0],
        )

        # Get end_time input value
        self.end_time = valid_time(self.pipeline.get(
            'end_time', default=[None])[0],
        )

        # get a database session
        self.session = Connector.manager[PIPELINE_DATABASE].session

        # Get tryouts from pipeline properties
        self.tryouts = self.pipeline.get(
            'tryouts', default=[TRYOUTS], create=True)[0]

        # Get wait from pipeline properties
        self.wait = self.pipeline.get(
            'wait', default=[TIME_WAIT_SEC], create=True)[0]

        # Retrieve --limit keyword value
        self.limit = self.pipeline.get('limit',
                                       default=[SQL_LIMIT],
                                       args=True)[0]

        # Database insertion datetime
        self.insert_time = datetime.today()

    def run(self):
        # Define task job ID (long and short)
        self.job_uuid = str(uuid.uuid4())
        self.job_id = f'UpdateUtcTime-{self.job_uuid[:8]}'
        logger.info(f'Task {self.job_id} is starting')
        try:
            self.setup_inputs()
        except:
            logger.exception(f'Initializing inputs has failed for task {self.job_id}!')
            self.pipeline.exit()
            return

        # List of tables for which utc times must be updated
        table_list = [TmLog, SbmLog, BiaSweepLog, LfrKcoeffDump]

        # Loop over tables
        for current_table in table_list:
            # Retrieve table rows which are flagged as predictive in the table
            filters = [current_table.utc_time_is_predictive == True]
            if self.start_time:
                filters.append(current_table.utc_time >= str(self.start_time))
            if self.end_time:
                filters.append(current_table.utc_time < str(self.end_time))
            filters = and_(*filters)
            rows = query_db(self.session, current_table,
                            filters=filters,
                            tryouts=self.tryouts,
                            wait=self.wait,
                            limit=self.limit,
                            to_dict='records')
            n_rows = len(rows)
            logger.info(f'{n_rows} predictive time for {current_table.__tablename__} [{self.job_id}]')
            if n_rows == 0:
                continue

            # Check/Update utc time
            updated_rows = [current_row
                            for current_row in rows
                            if self.update_time(current_row)]
            n_rows = len(updated_rows)

            if n_rows > 0:
                # Then commit changes in database
                self.session.commit()
                logger.info(f'{n_rows} updated in {current_table.__tablename__} [{self.job_id}]')
            else:
                logger.info(f'No data to update in {current_table.__tablename__} [{self.job_id}]')

    def update_time(self, row, naif_id=NAIF_SOLO_ID):
        """
        Check if utc_time in the input row is up to date.
        If not, then update it using SPICE.

        :param row: input table row to check/update (dictionary)
        :return:True if row has been updated, False otherwise
        """
        is_uptodate = False
        if is_sclk_uptodate(row['utc_time'] + timedelta(days=1), Path(self.sclk_file).name):
            row['utc_time'] = self.spice.obt2utc(naif_id, row['cuc_time'])
            row['utc_time_is_predictive'] = False
            row['insert_time'] = self.insert_time
            logger.debug(f'{row} has been updated [{self.job_id}]')
            is_uptodate = True

        return is_uptodate



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['roc',
 'roc.dingo',
 'roc.dingo.models',
 'roc.dingo.models.versions',
 'roc.dingo.tasks',
 'roc.dingo.templates',
 'roc.dingo.tests',
 'roc.dingo.tests.test-1-tasks',
 'roc.dingo.tests.test-2-commands']

package_data = \
{'': ['*'],
 'roc.dingo.models.versions': ['SQL/*'],
 'roc.dingo.tests': ['data-tasks/*']}

install_requires = \
['SQLAlchemy>=1.3,<2.0',
 'alembic>=1.4,<2.0',
 'astropy>=5.2,<6.0',
 'numpy!=1.19.5',
 'pandas>=1.1.3',
 'poppy-core>=0.9.4',
 'poppy-pop>=0.7.5',
 'psycopg2>=2.8.4,<3.0.0',
 'roc-idb>=1.0,<2.0',
 'roc-rpl>=1.0,<2.0',
 'spacepy>=0.4.1,<0.5.0',
 'spice_manager>=1.1.0,<2.0.0',
 'xmltodict>=0.13,<0.14']

setup_kwargs = {
    'name': 'roc-dingo',
    'version': '1.3.6',
    'description': 'Data INgestOr (DINGO) plugin is used to ingest data into the ROC pipeline database',
    'long_description': 'DINGO PLUGIN README\n===================\n\n[![pipeline status](https://gitlab.obspm.fr/ROC/Pipelines/Plugins/DINGO/badges/develop/pipeline.svg)](https://gitlab.obspm.fr/ROC/Pipelines/Plugins/DINGO/pipelines)\n[![coverage report](https://gitlab.obspm.fr/ROC/Pipelines/Plugins/DINGO/badges/develop/coverage.svg)](https://roc.pages.obspm.fr/Pipelines/Plugins/DINGO/pytest_coverage_html/)\n[![tests status](https://roc.pages.obspm.fr/Pipelines/Plugins/DINGO/pie.svg)](https://roc.pages.obspm.fr/Pipelines/Plugins/DINGO/report.html)\n\nThis directory contains the source files of the Data INGestOr (DINGO), a plugin of the ROC pipeline used to ingest data into the ROC database.\nDINGO is developed with and run under the POPPY framework.\n\n## User guide\n\n### Pre-requisites\n\nThe following software must be installed:\n- Python 3.8\n- pip tool\n- poetry (optional)\n- git (optional)\n\n### Install a stable release with pip\n\nTo install the roc-dingo plugin with pip:\n\n``pip install roc-dingo``\n\n## Nominal usage\n\nroc-dingo is designed to be called from a pipeline running with the POPPy framework.\n\nThe plugin can be used in Python programs using "import roc.dingo".\n\n## Developer guide\n\n### Install a local copy from source files\n\nTo install a local copy of the roc-dingo plugin:\n\n1. Retrieve a copy of the source files from https://gitlab.obspm.fr/ROC/Pipelines/Plugins/DINGO (restricted access)\n2. Use `pip install` or `poetry install` command to install local instance\n\n### Publish a new version\n\n1. Update the plugin version in pyproject.toml\n2. Update the plugin descriptor using ``python bump_descriptor.py -m <message>``\n3. Update `poetry.lock` file running `poetry lock`\n4. Always commit in `develop` branch first\n5. Merge in `master/main` branch and tag the version\n\nN.B. When a new tag is pushed in gitlab, tests are automatically run in pipeline, then plugin published in pypi.\n\nAuthors\n-------\n\n* Florence HENRY florence.henry@obspm.fr (maintainer)\n* Xavier BONNIN xavier.bonnin@obspm.fr (maintainer)\n* Sonny LION sonny.lion@obspm.fr (author)\n\nLicense\n-------\n\nThis project is licensed under CeCILL-C.\n\nAcknowledgments\n---------------\n\n* Solar Orbiter / RPW Operation Centre (ROC) team\n',
    'author': 'Florence Henry',
    'author_email': 'florence.henry@obspm.fr',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.obspm.fr/ROC/Pipelines/Plugins/DINGO',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4',
}


setup(**setup_kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Reference documentation: ROC-GEN-SYS-NTT-00038-LES

"""
Database model for tm_log and tc_log tables.
"""

from poppy.core.db.non_null_column import NonNullColumn

from poppy.core.db.base import Base
from sqlalchemy import String, ForeignKey, UniqueConstraint, Index
from sqlalchemy.orm import relationship, validates, backref
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.dialects.postgresql import (
    BIGINT,
    BOOLEAN,
    DOUBLE_PRECISION,
    ENUM,
    INTEGER,
    SMALLINT,
    TIMESTAMP,
    JSONB,
)

__all__ = [
    'TmLog',
    'TcLog',
    'InvalidPacketLog',
]


class TmLog(Base):
    """
    Class representation of the table for tm_log table in the ROC database.
    """

    id = NonNullColumn(BIGINT(), primary_key=True)
    length = NonNullColumn(INTEGER(), nullable=True,
                           descr='Packet length in bytes')
    category = NonNullColumn(String(512), nullable=True,
                             descr='Packet PALISADE category')
    apid = NonNullColumn(INTEGER(), nullable=True,
                         descr='Packet APID')
    sync_flag = NonNullColumn(BOOLEAN, nullable=True,
                              descr='TM packet time synchronization flag')
    srdb_id = NonNullColumn(String(16), nullable=True,
                            descr='Packet name (SRDB ID)')
    palisade_id = NonNullColumn(String(256), nullable=True,
                                descr='Packet PALISADE ID')
    binary = NonNullColumn(String(), nullable=True,
                           descr='Packet raw binary data (in hexadecimal)')
    data = NonNullColumn(JSONB(), nullable=True,
                           descr='Packet source data (JSON format)')
    sha = NonNullColumn(String(),
                        descr='Packet sha (hexdigest)')
    sequence_cnt = NonNullColumn(BIGINT(), nullable=True,
                                 descr='Packet sequence counter')
    cuc_time = NonNullColumn(String(1024),
                             descr='Packet creation on-board time in CCSDS CUC format "coarse:fine".')
    obt_time = NonNullColumn(TIMESTAMP, nullable=True,
                             descr='Packet creation on-board time in SQL timestamp format".')
    utc_time = NonNullColumn(TIMESTAMP, nullable=True,
                             descr='Packet creation UTC time in SQL timestamp format".')
    utc_time_is_predictive = NonNullColumn(BOOLEAN, nullable=True,
                                           descr='Flag to indicate if Packet creation UTC time is predictive (True) or definitive (False)')
    insert_time = NonNullColumn(TIMESTAMP,
                                descr='Database insertion local time.')

    __tablename__ = 'tm_log'
    __table_args__ = (
        UniqueConstraint('sha'),
        {
            'schema': 'pipeline',
        }
    )


class TcLog(Base):
    """
    Class representation of the table for tc_log table in the ROC database.
    """

    id = NonNullColumn(BIGINT(), primary_key=True)
    length = NonNullColumn(INTEGER(),
                           descr='Packet length in bytes')
    category = NonNullColumn(String(512), nullable=True,
                             descr='Packet PALISADE category')
    apid = NonNullColumn(INTEGER(), nullable=True,
                         descr='Packet APID')
    utc_time = NonNullColumn(TIMESTAMP,
                             descr='Packet execution time in UTC')
    srdb_id = NonNullColumn(String(16),
                            descr='Packet name (SRDB ID)')
    palisade_id = NonNullColumn(String(256),
                                descr='Packet PALISADE ID')
    binary = NonNullColumn(String(), nullable=True,
                           descr='Packet raw binary data (in hexadecimal)')
    data = NonNullColumn(JSONB(), nullable=True,
                           descr='Packet application data (JSON format)')
    sha = NonNullColumn(String(),
                        descr='Packet sha (hexdigest)')
    tc_exe_state = NonNullColumn(String(16), nullable=True,
                                 descr='TC acknowledgment'
                                 'execution completion status')
    tc_acc_state = NonNullColumn(String(16), nullable=True,
                                 descr='TC acknowledgment acceptance status')
    sequence_name = NonNullColumn(String(16), nullable=True,
                                  descr='TC sequence name')
    unique_id = NonNullColumn(String(256), nullable=True,
                              descr='TC unique ID')
    insert_time = NonNullColumn(TIMESTAMP,
                                descr='Database insertion local time.')

    __tablename__ = 'tc_log'
    __table_args__ = (
        UniqueConstraint('sha'),
        {
            'schema': 'pipeline',
        }
    )


class InvalidPacketLog(Base):
    """
    Class representation of the table for invalid_packet_log table in the ROC database.
    """

    id = NonNullColumn(BIGINT(), primary_key=True)
    sha = NonNullColumn(String(),
                        descr='sha256 computed from invalid packet data (hexdigest)')
    srdb_id = NonNullColumn(String(16), nullable=True,
                            descr='Packet name (SRDB ID)')
    apid = NonNullColumn(INTEGER(), nullable=True,
                         descr='Packet APID')
    palisade_id = NonNullColumn(String(256), nullable=True,
                                descr='Packet PALISADE ID')
    utc_time = NonNullColumn(TIMESTAMP, nullable=True,
                             descr='Packet time in UTC')
    data = NonNullColumn(JSONB(), nullable=True,
                           descr='Packet content data (when extraction is possible)')
    status = NonNullColumn(SMALLINT(),
                           descr='Status of the packet (INVALID_PACKET_HEADER = 1, '
                                 'INVALID_PACKET_DATA = 2, OUTSIDE_RANGE_PACKET = 3, '
                                 'INVALID_PACKET_CDATA = 4, INVALID_PACKET_TIME = 5, '
                                 'UNKNOWN_STATUS = -1)')
    comment = NonNullColumn(String(), nullable=True,
                        descr='Additional comment about packet status')
    insert_time = NonNullColumn(TIMESTAMP,
                                descr='Database insertion local time.')

    __tablename__ = 'invalid_packet_log'
    __table_args__ = (
        UniqueConstraint('sha'),
        {
            'schema': 'pipeline',
        }
    )



"""empty message

Revision ID: roc_dingo_0021_add_retrieved_time
Revises: roc_dingo_0020_rm_biacurrtablehfr
Create Date: 2022-01-17 15:26:10.806747

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

from poppy.pop.alembic.helpers import \
    create_table, execute, user, drop_table

# revision identifiers, used by Alembic.
revision = 'roc_dingo_0021_add_retrieved_time'
down_revision = 'roc_dingo_0020_rm_biacurrtablehfr'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('sbm_log', sa.Column('retrieved_time', postgresql.TIMESTAMP(), nullable=True), schema='pipeline')

    # added by hand
    # fix table permissions
    grant_user_seq = (
        """GRANT USAGE, SELECT
        ON ALL SEQUENCES IN SCHEMA pipeline TO {0}""".format(user)
    )
    execute(grant_user_seq)

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('sbm_log', 'retrieved_time', schema='pipeline')
    # ### end Alembic commands ###



# -*- coding: utf-8 -*-
"""Rename column type to level

Revision ID: 0005
Revises: roc_dingo_0004_modify_file_log
Create Date: 2020-10-14 14:37:49.295678

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = 'roc_dingo_0005_mv_type_level'
down_revision = 'roc_dingo_0004_modify_file_log'
depends_on = None


def upgrade():
    op.execute('DROP TYPE IF EXISTS file_level_list')
    op.execute('ALTER TYPE file_type_list  RENAME TO file_level_list')

    # rename type -> level
    op.alter_column(
        'file_log', 'type',
        new_column_name='level', schema='pipeline')


def downgrade():

    op.execute('ALTER TYPE file_level_list RENAME TO file_type_list')

    # rename level -> type
    op.alter_column(
        'file_log', 'level',
        new_column_name='type', schema='pipeline')
